from __future__ import annotations
from typing import Callable, Generator, Generic, IO, Iterable, Optional, Sequence, Tuple, TypeVar, cast, Hashable, Union
import types
from dataclasses import dataclass, field

#--------------------------------------------------
# FrozenOrderedSet
#--------------------------------------------------

T = TypeVar('T')
class FrozenOrderedSet(Generic[T]):
    """ Immutable access to an ordered sequence of elements without duplicates. """

    def __init__(self, data:Sequence[T]):
        # TODO - maybe verify that there are no duplicates?
        self.data = tuple(data)

    @classmethod
    def from_iterable(cls, items:Iterable[T]|None):
        return OrderedSet.from_iterable(items).frozen()

    def some(self) -> T:
        assert len(self.data) > 0
        return self.data[0]

    def includes(self, other: Iterable[T]) -> bool:
        """ True iff all items in other are in this set. """
        for x in other:
            if x not in self:
                return False
        return True

    def __hash__(self) -> int:
        return hash(self.data)

    def __eq__(self, other):
        if not isinstance(other, FrozenOrderedSet):
            return False
        return self.data == other.data

    def __getitem__(self, ix):
        if len(self.data) <= ix:
            return None
        return self.data[ix]

    def __contains__(self, item:T):
        return item in self.data

    def __bool__(self):
        return bool(self.data)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __sub__(self, other:Optional[Iterable[T]]) -> FrozenOrderedSet[T]:
        if not other:
            return self
        # set difference guaranteeing deterministic order
        new = OrderedSet[T]()
        new.update(self.data)
        new.difference_update(other)
        return new.frozen()

    def __or__(self, other:Optional[Iterable[T]]) -> FrozenOrderedSet[T]:
        if not other:
            return self
        # set union guaranteeing deterministic order
        new = OrderedSet[T]()
        new.update(self.data)
        new.update(other)
        return new.frozen()

    def __and__(self, other:Optional[Iterable[T]]) -> FrozenOrderedSet[T]:
        if not other:
            # intersetion is empty
            return frozen()
        # set intersection guaranteeing deterministic order
        new = OrderedSet[T]()
        for item in self:
            if item in other:
                new.add(item)
        return new.frozen()

    def __str__(self) -> str:
        return self.data.__str__()

#--------------------------------------------------
# OrderedSet
#--------------------------------------------------

T = TypeVar('T')
class OrderedSet(Generic[T]):
    def __init__(self):
        self.set:Optional[set[T]] = None
        self.list:Optional[list[T]] = None

    @classmethod
    def from_iterable(cls, items:Iterable[T]|None):
        if not items:
            return OrderedSet()
        s = OrderedSet()
        s.update(items)
        return s

    def _ensure_initialized(self):
        if self.set is None:
            self.set = set()
            self.list = []

    def get_set(self) -> set:
        self._ensure_initialized()
        assert(self.set is not None)
        return self.set

    def get_list(self) -> list:
        self._ensure_initialized()
        assert(self.list is not None)
        return self.list

    def add(self, item:T|None):
        """Insert an item at the end of the set."""
        if item not in self.get_set():
            self.get_set().add(item)
            self.get_list().append(item)
        return self

    def prepend(self, item:T|None):
        """Insert an item at the start of the set."""
        if item not in self.get_set():
            self.get_set().add(item)
            self.get_list().insert(0, item)
        return self

    def update(self, items:Iterable[T]|None):
        """Insert items at the end of the set."""
        if items is not None:
            for item in items:
                self.add(item)
        return self

    def prefix(self, items:Iterable[T]|None):
        """Insert items at the start of the set."""
        if items is not None:
            for item in items:
                self.prepend(item)
        return self

    def remove(self, item:T):
        if self.set is not None and item in self.set:
            self.difference_update([item])

    def difference_update(self, items:Iterable[T]):
        if self.set is None:
            return

        changed = False
        for item in items:
            if item in self.set:
                self.set.remove(item)
                changed = True

        if changed:
            # list.remove uses == under the covers, which screws
            # with our DSL objects that have __eq__ defined and
            # return expressions, so we need to do this manually
            new_list = []
            for cur in self.get_list():
                if cur in self.set:
                    new_list.append(cur)
            self.list = new_list

    def clear(self):
        if self.set is not None:
            self.set.clear()
        if self.list is not None:
            self.list.clear()

    def pop(self) -> T:
        item = self.get_list().pop()
        self.get_set().remove(item)
        return item

    def first(self) -> T:
        assert len(self.get_list()) > 0
        return self.get_list()[0]
    
    def some(self) -> T:
        return self.first()

    def frozen(self) -> FrozenOrderedSet[T]:
        return FrozenOrderedSet(self.get_list())

    def version(self) -> tuple[int, T|None]:
        return (len(self.list) if self.list else 0, self.list[-1] if self.list and len(self.list) > 0 else None)

    def has_changed(self, version:tuple[int, T|None]) -> bool:
        if version[0] != len(self.get_list()):
            return True
        if version[1] is None:
            return len(self.get_list()) > 0
        if version[1] is not self.get_list()[-1]:
            return True
        return False

    def includes(self, other: Iterable[T]) -> bool:
        """ True iff all items in other are in this set. """
        for x in other:
            if x not in self:
                return False
        return True

    def __hash__(self) -> int:
        return hash((tuple(self.get_list())))

    def __contains__(self, item:T):
        return self.set and item in self.set

    def __bool__(self):
        return bool(self.set)

    def __getitem__(self, ix) -> T:
        if not self.list or ix >= len(self.get_list()):
            raise IndexError
        return self.get_list()[ix]

    def __iter__(self):
        return iter(self.list) if self.list else iter([])

    def __len__(self):
        return len(self.list) if self.list is not None else 0

    def __sub__(self, other:Optional[Iterable[T]]) -> OrderedSet[T]:
        if not other:
            return self
        # set difference guaranteeing deterministic order
        new = OrderedSet[T]()
        new.update(self)
        for item in other:
            new.remove(item)
        return new

    def __and__(self, other: Optional[Iterable[T]]) -> OrderedSet[T]:
        if not other:
            # intersetion is empty
            return ordered_set()
        # set intersection guaranteeing deterministic order
        new = OrderedSet[T]()
        for item in self:
            if item in other:
                new.add(item)
        return new

    def __or__(self, other: Optional[Iterable[T]]) -> OrderedSet[T]:
        if not other:
            return self
        # set union guaranteeing deterministic order
        new = OrderedSet[T]()
        new.update(self)
        new.update(other)
        return new

    def __isub__(self, other: Iterable[T]) -> OrderedSet[T]:
        for item in other:
            self.remove(item)
        return self

    def __ior__(self, other: Iterable[T]) -> OrderedSet[T]:
        self.update(other)
        return self

    def __iand__(self, other: Iterable[T]) -> OrderedSet[T]:
        for item in other:
            if item not in self:
                self.remove(item)
        return self

    def __str__(self) -> str:
        return self.list.__str__()


T = TypeVar('T')
def ordered_set(*items: T) -> OrderedSet[T]:
    """ Create an OrderedSet with these items. """
    s = OrderedSet()
    if items is not None:
        s.update(items)
    return s

def frozen(*items: T) -> FrozenOrderedSet[T]:
    """ Create a FrozenOrderedSet with these items."""
    return FrozenOrderedSet(items)

V = TypeVar('V')
K = TypeVar('K')
def index_by(s: Iterable[V], f:Callable[[V], K]) -> dict[K, V]:
    """ Create an index for the sequence by computing a key for each value using this function. """
    d = dict()
    for v in s:
        d[f(v)] = v
    return d

V = TypeVar('V')
K = TypeVar('K')
def group_by(s: Iterable[V], f:Callable[[V], K]) -> dict[K, OrderedSet[V]]:
    """ Group elements of the sequence by a key computed for each value using this function. """
    d = dict()
    for v in s:
        key = f(v)
        if key not in d:
            d[key] = OrderedSet()
        d[key].add(v)
    return d

def split_by(s: Iterable[V], f:Callable[[V], bool]) -> Tuple[list[V], list[V]]:
    """ Split the iterable in 2 groups depending on the result of the callable: [True, False]."""
    trues = []
    falses = []
    for v in s:
        trues.append(v) if f(v) else falses.append(v)
    return (trues, falses)

def filter_by_type(s: Iterable[V], types)-> list[V]:
    """ Filter the iterable keeping only elements of these types. """
    r = []
    for v in s:
        if isinstance(v, types):
            r.append(v)
    return r

def rewrite_set(t: type[T], f: Callable[[T], T], items: FrozenOrderedSet[T]) -> FrozenOrderedSet[T]:
    """ Map a function over a set, returning a new set with the results. Avoid allocating a new set if the function is the identity. """
    new_items: Optional[list[T]] = None
    for i in range(len(items)):
        item = items[i]
        assert isinstance(item, t), f"Expected {t}, got {type(item)}" # shut up type checker
        new_item = cast(T, f(item))
        assert isinstance(new_item, t), f"Expected {t}, got {type(new_item)}" # shut up type checker
        if new_item is not item:
            if new_items is None:
                new_items = cast(list[T], list(items))
            new_items[i] = new_item
    if new_items is None:
        return items
    return ordered_set(*new_items).frozen()

def rewrite_list(t: Union[type[T], types.UnionType], f: Callable[[T], T], items: Tuple[T, ...]) -> Tuple[T, ...]:
    """ Map a function over a list, returning a new list with the results. Avoid allocating a new list if the function is the identity. """
    new_items: Optional[list[T]] = None
    for i in range(len(items)):
        item = items[i]
        new_item = cast(T, f(item))
        if new_item is not item:
            if new_items is None:
                new_items = cast(list[T], list(items))
            new_items[i] = new_item
    if new_items is None:
        return items
    return tuple(new_items)

def flatten_iter(items: Iterable[object], t: Union[type[T], types.UnionType]) -> Generator[T, None, None]:
    """Yield items from a nested iterable structure one at a time."""
    for item in items:
        if isinstance(item, (list, tuple, OrderedSet)):
            yield from flatten_iter(item, t)
        elif isinstance(item, t):
            yield cast(T, item)

def flatten_tuple(items: Iterable[object], t: Union[type[T], types.UnionType]) -> tuple[T, ...]:
    """ Flatten the nested iterable structure into a tuple."""
    return tuple(flatten_iter(items, t))

@dataclass(frozen=True)
class NameCache:
    # support to generate object names with a count when there's collision
    # the next count to use as a suffix for an object with this name
    name_next_count: dict[Hashable, dict[str, int]] = field(default_factory=dict)
    # cache of the precomputed name for the object with this key
    name_cache: dict[Hashable, str] =  field(default_factory=dict)
    # whether to use _ or not
    use_underscore: bool = True
    # whether the first entry should start with a 1 or with nothing
    start_from_one: bool = False

    def get_name(self, key: Hashable, name: str, prefix: str = "") -> str:
        """
        Generate a unique name for the given key and base name, avoiding name collisions.

        Names are tracked and deduplicated using an internal counter. If a name has already
        been generated for the given key, the cached name is returned.

        For tuple keys (e.g., (relation_id, var_id)), the first element is used as the scope
        for counting name collisions. Only tuples of length 2 are supported; an assertion will
        fail otherwise.

        Parameters:
            key (Hashable): A unique key identifying the object. Can be an int or a (scope_id, local_id) tuple.
            name (str): Base name to use.
            prefix (str): Optional prefix to prepend to the name.

        Returns:
            str: A unique name string for the given key.

        Examples:
            # Global naming
            get_name(1, "var")           => "var"
            get_name(2, "var")           => "var_2"
            get_name(3, "var")           => "var_3"

            # Scoped naming (e.g., per relation)
            get_name((10, 1), "x")       => "x"
            get_name((10, 2), "x")       => "x_2"
            get_name((20, 1), "x")       => "x"     # Different scope, starts again
            get_name((20, 2), "x")       => "x_2"

            # With prefix and custom suffix start
            n = NameCache()
            n.get_name(1, "val", prefix="t_")                => "t_val"

            n = NameCache(start_from_one=True)
            n.get_name(1, "val", prefix="t_")                => "t_val_1"
        """
        if key in self.name_cache:
            return self.name_cache[key]

        # Derive the count scope from the key
        scope = None
        if isinstance(key, tuple):
            assert len(key) == 2, f"Expected tuple key of length 2, got {len(key)}: {key}"
            scope = key[0]  # e.g., relation_id

        # get the dict specific for the scope, or create one
        if scope in self.name_next_count:
            next_count = self.name_next_count[scope]
        else:
            next_count = dict()
            self.name_next_count[scope] = next_count

        # find the next available name
        name = self._find_next(f"{prefix}{name}", next_count)

        # register it for the key
        self.name_cache[key] = name
        return name

    def _find_next(self, name: str, next_count: dict):
        if name in next_count:
            # name is already in use, so append the next count
            c = next_count[name]
            next_count[name] = c + 1
            new_name = self._concat(name, c)

        else:
            # name not in use yet, so register it
            next_count[name] = 2
            if self.start_from_one:
                new_name = self._concat(name, 1)
            else:
                new_name = name

        if new_name != name:
            if new_name in next_count:
                # if the modified name is also in use, recurse
                return self._find_next(name, next_count)
            else:
                # otherwise, record that the modified name was used
                next_count[new_name] = 2
        return new_name

    def _concat(self, name: str, c: int):
        return f"{name}_{c}" if self.use_underscore else f"{name}{c}"

@dataclass(frozen=True)
class Printer():
    """ Helper class to print ASTs into an IO object. """
    io: Optional[IO[str]] = None

    def _nl(self) -> None:
        """ Print a new line character. """
        self._print("\n")

    def _print(self, arg) -> None:
        """ Print the argument into the io without indentation. """
        if self.io is None:
            print(arg, end='')
        else:
            self.io.write(arg)

    def _print_nl(self, arg) -> None:
        """ Helper for the common case of calling _print followed by _nl. """
        self._print(arg)
        self._nl()

    def _indent_print(self, depth, arg) -> None:
        """ Print the argument into the io with indentation based on depth. """
        if self.io is None:
            print("    " * depth + str(arg), end='')
        else:
            self.io.write("    " * depth + str(arg))

    def _indent_print_nl(self, depth, arg) -> None:
        """ Helper for the common case of calling _indent_print followed by _nl. """
        self._indent_print(depth, arg)
        self._nl()



from urllib.parse import urlparse

# keyring is not available in the snowflake notebook environment
try:
    import keyring
    KEYRING_AVAILABLE = True
except ImportError:
    KEYRING_AVAILABLE = False
    keyring = None

def is_localhost(url: str) -> bool:
    parsed = urlparse(url)
    hostname = parsed.hostname

    return hostname in {"localhost", "127.0.0.1", "::1"}

def extract_port(url: str) -> int:
    parsed = urlparse(url)
    port = parsed.port
    if port is None:
        if parsed.scheme == "https":
            port = 443
        else:
            port = 80
    return port

def remove_url_scheme(url: str) -> str:
    parsed = urlparse(url)
    # If there's a scheme, reconstruct without it
    if parsed.scheme:
        return parsed.netloc + parsed.path
    return url
    
def is_keyring_secure() -> bool:
    if not KEYRING_AVAILABLE or keyring is None:
        return False
    kr = keyring.get_keyring()
    if kr.__class__.__name__ in ("PlaintextKeyring", "fail"):
        print("Warning: Keyring is not secure. Please use a secure keyring backend.")
        return False
    return True



"""
Converter for DBT profiles.yml to RAI Config format.
"""

from __future__ import annotations

import os
import re
from typing import Any

from .dbt_models import DBTProfile, DBTProfilesFile


def render_jinja_value(value: str) -> str:
    """
    Render Jinja2 env_var() templates in DBT profiles.

    Supports: {{ env_var('VAR_NAME') }} or {{ env_var("VAR_NAME") }}
    """
    pattern = r"\{\{\s*env_var\(['\"]([^'\"]+)['\"]\)\s*\}\}"

    def replace_env_var(match):
        var_name = match.group(1)
        env_value = os.environ.get(var_name)
        if env_value is None:
            raise ValueError(
                f"Environment variable '{var_name}' referenced in profiles.yml is not set"
            )
        return env_value

    return re.sub(pattern, replace_env_var, value)


def process_value(value: Any) -> Any:
    if isinstance(value, str):
        return render_jinja_value(value)
    elif isinstance(value, dict):
        return {k: process_value(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [process_value(v) for v in value]
    else:
        return value


def convert_dbt_to_rai(
    dbt_profiles: dict[str, Any],
    profile_name: str | None = None,
    target_name: str | None = None
) -> dict[str, Any]:
    profiles_file = DBTProfilesFile(**dbt_profiles)
    dbt_profiles = profiles_file.model_dump()

    dbt_profiles = process_value(dbt_profiles)

    if not profile_name:
        profile_name = os.environ.get("DBT_PROFILE")

    if not profile_name:
        profile_names = [k for k in dbt_profiles.keys() if k != "config" and not k.startswith("config_")]
        if not profile_names:
            raise ValueError("No profiles found in profiles.yml")
        profile_name = profile_names[0]

    if profile_name not in dbt_profiles:
        available = [k for k in dbt_profiles.keys() if k != "config"]
        raise ValueError(
            f"Profile '{profile_name}' not found. Available profiles: {available}"
        )

    profile_data = dbt_profiles[profile_name]
    profile = DBTProfile(**profile_data)

    # outputs existence validated by DBTProfile model
    # Get target (from parameter, env var, target field, or 'dev')
    if not target_name:
        target_name = os.environ.get("DBT_TARGET")
    if not target_name:
        target_name = profile.target
    if not target_name:
        # Try 'dev' first, then first output
        if "dev" in profile.outputs:
            target_name = "dev"
        else:
            target_name = next(iter(profile.outputs.keys()))

    if target_name not in profile.outputs:
        available = list(profile.outputs.keys())
        raise ValueError(
            f"Target '{target_name}' not found in profile '{profile_name}'. "
            f"Available targets: {available}"
        )

    output = profile.outputs[target_name]

    connection = output.convert()
    connection_name = output.provider

    return {
        "connections": {connection_name: connection},
        "default_connection": connection_name,
    }



from __future__ import annotations

from . import StringValue, IntegerValue, DateValue, DateTimeValue, math
from ..frontend.base import Aggregate, Library, Concept, NumberConcept, Expression, Field, Literal, TupleVariable, Variable, Value
from ..frontend.core import Any, Int128, Number, String, Integer, Date, DateTime, Hash

from typing import Union, Literal
import datetime as dt

# the front-end library object
library = Library("common")


_range = library.Relation("range", [Field.input("start", Integer), Field.input("stop", Integer), Field.input("step", Integer), Field("value", Integer)])

def range(*args: IntegerValue) -> Expression:
    """ Generate a range of integers.

    Supports range(stop), range(start, stop), range(start, stop, step).

    Start is inclusive and defaults to 0.
    Stop is exclusive.
    Step defaults to 1.
    """
    if len(args) == 1:
        start, stop, step = 0, args[0], 1
    elif len(args) == 2:
        start, stop, step = args[0], args[1], 1
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise ValueError(f"range expects 1, 2, or 3 arguments, got {len(args)}")

    return _range(start, stop, step)

_hash = library.Relation("hash", [Field.input("args", Any, is_list=True), Field("hash", Hash)])
def hash(*args: Value) -> Expression:
    """ Compute a hash value for the given arguments. """
    return _hash(TupleVariable(args))

_uuid_to_string = library.Relation("uuid_to_string", [Field.input("uuid", Hash), Field("str", String)])
def uuid_to_string(uuid: Value) -> Expression:
    """ Convert a UUID (Hash) to its string representation. """
    return _uuid_to_string(uuid)


"""Minimal entry point for the ``rai-dev`` CLI."""

from __future__ import annotations

import click

VERSION = "1.0.0a1"


@click.command()
@click.version_option(version=VERSION, prog_name="rai-dev")
def cli() -> None:
    """Display the CLI version."""
    click.echo(f"rai-dev version {VERSION}")


if __name__ == "__main__":
    cli()




from __future__ import annotations
from typing import Union, Sequence

from v0.relationalai.semantics.internal import internal as b
from .std import _String, _Integer, _make_expr

# TODO(coey) can we allow multiple args to this, and convert each to string and concatenate? similar to e.g. julia `string("foo", 1, 'b')`. then we don't need a separate concat function (and concat only works on strings, which is inconvenient).
def string(s: Union[b.Producer, str, float, int]) -> b.Expression:
    return _make_expr("string", s, b.String.ref("res"))

def concat(s0: _String, s1: _String, *args: _String) -> b.Expression:
    res = _make_expr("concat", s0, s1, b.String.ref("res0"))
    for i, s in enumerate(args):
        res = _make_expr("concat", res, s, b.String.ref(f"res{i+1}"))
    return res

def len(s: _String) -> b.Expression:
    return _make_expr("num_chars", s, b.Int64.ref("res"))

def startswith(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("starts_with", s0, s1)

def endswith(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("ends_with", s0, s1)

def contains(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("contains", s0, s1)

def substring(s: _String, start: _Integer, stop: _Integer) -> b.Expression:
    # unlike Python, Rel's range is 1..stop inclusive
    return _make_expr("substring", s, start+1, stop, b.String.ref("res"))

def like(s: _String, pattern: _String) -> b.Expression:
    return _make_expr("like_match", s, pattern)

def lower(s: _String) -> b.Expression:
    return _make_expr("lower", s, b.String.ref("res"))

def upper(s: _String) -> b.Expression:
    return _make_expr("upper", s, b.String.ref("res"))

def strip(s: _String) -> b.Expression:
    return _make_expr("strip", s, b.String.ref("res"))

def levenshtein(s: _String, t: _String) -> b.Expression:
    return _make_expr("levenshtein", s, t, b.Int64.ref("res"))

def join(strs: Sequence[_String], separator: _String) -> b.Expression:
    return _make_expr("join", b.TupleArg(strs), separator, b.String.ref("res"))

def replace(s: _String, old: _String, new: _String) -> b.Expression:
    return _make_expr("replace", s, old, new, b.String.ref("res"))

def split(s: _String, separator: _String) -> tuple[b.Producer, b.Producer]:
    exp = _make_expr("split", separator, s, b.Int64.ref("idx"), b.String.ref("res"))
    # indexes are 0-based
    return exp._arg_ref(2) - 1, exp._arg_ref(3)

def split_part(s: _String, separator: _String, idx: _Integer) -> b.Expression:
    return _make_expr("split_part", separator, s, idx + 1, b.String.ref("res"))

def regex_match(s:_String, regex: _String) -> b.Expression:
    return _make_expr("regex_match", regex, s)


import dataclasses
from typing import Optional

from v0.relationalai.semantics.internal import internal as builder
from v0.relationalai.semantics import Float, Int128, String, Bool, Date, DateTime, Concept as QBConcept

#=
# Physical metadata for a Snowflake table.
#=

@dataclasses.dataclass(frozen=True)
class ColumnRef:
    table: str
    column: str

@dataclasses.dataclass
class ForeignKey:
    name: str
    source_columns: list[ColumnRef] = dataclasses.field(default_factory=list)
    target_columns: list[ColumnRef] = dataclasses.field(default_factory=list)

    def __hash__(self):
        return hash(self.name)

@dataclasses.dataclass
class SchemaMetadata:
    name: str
    foreign_keys: list[ForeignKey] = dataclasses.field(default_factory=list)

@dataclasses.dataclass
class ColumnMetadata:
    name: str
    datatype: str
    is_nullable: bool
    numeric_precision: Optional[int] = None
    numeric_precision_radix: Optional[int] = None
    numeric_scale: Optional[int] = None

@dataclasses.dataclass
class CsvColumnMetadata:
    name: str
    datatype: QBConcept

@dataclasses.dataclass
class TabularMetadata:
    name: str
    columns: list[ColumnMetadata] = dataclasses.field(default_factory=list)
    foreign_keys: set[ForeignKey] = dataclasses.field(default_factory=set)


_sf_type_mapping = {
    'varchar': String,
    'char': String,
    'text': String,
    'date': Date,
    'datetime': DateTime,
    'timestamp_ntz': DateTime,
    'timestamp_tz': DateTime,
    'boolean': Bool,
    'float': Float,
}

def _map_rai_type(col: ColumnMetadata) -> QBConcept:
    datatype = col.datatype.lower()
    if datatype == 'number' or datatype == 'fixed':
        if col.numeric_scale is not None and col.numeric_scale > 0:
            return _map_decimal_type(col.numeric_precision, col.numeric_scale)
        else:
            return Int128
    else:
        return _sf_type_mapping[datatype]

def _map_decimal_type(precision: Optional[int], scale: Optional[int]) -> QBConcept:
    if precision is None or scale is None:
       raise ValueError("Size and scale must be provided for decimal type mapping")
    return builder.decimal_concept(precision, scale)



import warnings

from v0.relationalai.semantics.lqp.validators import assert_valid_input

__all__ = ["assert_valid_input"]

warnings.warn(
    "relationalai.early_access.lqp.validators is deprecated. "
    "Please migrate to relationalai.semantics.lqp.validators",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
from collections import defaultdict
from datetime import datetime, timezone
import atexit
import re

from pandas import DataFrame
from typing import Any, Optional, Literal, TYPE_CHECKING
from snowflake.snowpark import Session

from v0.relationalai import debugging
from v0.relationalai.errors import NonDefaultLQPSemanticsVersionWarning
from v0.relationalai.semantics.lqp import result_helpers
from v0.relationalai.semantics.metamodel import ir, factory as f, executor as e
from v0.relationalai.semantics.lqp.compiler import Compiler
from v0.relationalai.semantics.lqp.intrinsics import mk_intrinsic_datetime_now
from v0.relationalai.semantics.lqp.constructors import mk_transaction
from v0.relationalai.semantics.lqp.types import lqp_type_to_sql
from lqp import print as lqp_print, ir as lqp_ir
from lqp.parser import construct_configure
from v0.relationalai.semantics.lqp.ir import convert_transaction, validate_lqp
from v0.relationalai.clients.config import Config
from v0.relationalai.clients.resources.snowflake import APP_NAME, create_resources_instance
from v0.relationalai.clients.types import TransactionAsyncResponse
from v0.relationalai.clients.util import IdentityParser, escape_for_f_string
from v0.relationalai.tools.constants import QUERY_ATTRIBUTES_HEADER
from v0.relationalai.tools.query_utils import prepare_metadata_for_headers

if TYPE_CHECKING:
    from v0.relationalai.semantics.snowflake import Table

# Whenever the logic engine introduces a breaking change in behaviour, we bump this version
# once the client is ready to handle it.
DEFAULT_LQP_SEMANTICS_VERSION = "0"

class LQPExecutor(e.Executor):
    """Executes LQP using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        wide_outputs: bool = False,
        connection: Session | None = None,
        config: Config | None = None,
        # In order to facilitate snapshot testing, we allow overriding intrinsic definitions
        # like the current time, which would otherwise change between runs.
        intrinsic_overrides: dict = {},
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.wide_outputs = wide_outputs
        self.compiler = Compiler()
        self.connection = connection
        self.config = config or Config()
        self.intrinsic_overrides = intrinsic_overrides
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                # NOTE: language="lqp" is not strictly required for LQP execution, but it
                # will significantly improve performance.
                self._resources = create_resources_instance(
                    config=self.config,
                    dry_run=self.dry_run,
                    connection=self.connection,
                    language="lqp",
                )
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True, "lqp")
        return self._resources

    # Checks the graph index and updates it if necessary
    def prepare_data(self):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from v0.relationalai.semantics.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine
        engine_size = self.resources.config.get_default_engine_size()

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.maybe_poll_use_index(
                app_name=app_name,
                sources=sources,
                model=model,
                engine_name=self.engine,
                engine_size=engine_size,
                program_span_id=program_span_id,
            )

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from v0.relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _export(self, txn_id: str, export_info: tuple, dest: Table, actual_cols: list[str], declared_cols: list[str], update: bool):
        # At this point of the export, we assume that a CSV file has already been written
        # to the Snowflake Native App stage area. Thus, the purpose of this method is to
        # copy the data from the CSV file to the destination table.
        _exec = self.resources._exec
        dest_database, dest_schema, dest_table, _ = IdentityParser(dest._fqn, require_all_parts=True).to_list()
        filename = export_info[0]
        result_table_name = filename + "_table"

        with debugging.span("export", txn_id=txn_id, export_info=export_info, dest_table=dest_table):
            with debugging.span("export_to_result_schema"):
                # First, we need to persist from the CSV file to the results schema by calling the
                # `persist_from_stage` stored procedure. This step also cleans up the CSV file in
                # the stage area.
                column_fields = []
                for (col_name, col_type) in export_info[1]:
                    column_fields.append([col_name, lqp_type_to_sql(col_type)])

                # NOTE: the `str(column_fields)` depends on python formatting which surrounds
                # strings with single quotes. If this changes, or if we ever get a single quote in
                # the actual string, then we need to do something more sophisticated.
                exec_str = f"call {APP_NAME}.api.persist_from_stage('{txn_id}', '{filename}', '{result_table_name}', {str(column_fields)})"
                _exec(exec_str)

            with debugging.span("write_table"):
                # The result of the first step above is a table in the results schema,
                # {app_name}.results.{result_table_name}.
                # Second, we need to copy the data from the results schema to the
                # destination table. This step also cleans up the result table.
                out_sample = _exec(f"select * from {APP_NAME}.results.{result_table_name} limit 1;")
                names = self._build_projection(declared_cols, actual_cols, column_fields, out_sample)
                dest_fqn = dest._fqn
                try:
                    if not update:
                        createTableLogic = f"""
                                        CREATE TABLE {dest_fqn} AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{result_table_name};
                        """
                        if dest._is_iceberg:
                            assert dest._iceberg_config is not None
                            external_volume_clause = ""
                            if dest._iceberg_config.external_volume:
                                external_volume_clause = f"EXTERNAL_VOLUME = '{dest._iceberg_config.external_volume}'"
                            createTableLogic = f"""
                                        CREATE ICEBERG TABLE {dest_fqn}
                                        CATALOG = "SNOWFLAKE"
                                        {external_volume_clause}
                                        AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{result_table_name};
                            """

                        _exec(f"""
                            BEGIN
                                -- Check if table exists
                                IF (EXISTS (
                                    SELECT 1
                                    FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                    WHERE table_schema = '{dest_schema}'
                                    AND table_name = '{dest_table}'
                                )) THEN
                                    -- Insert into existing table
                                    EXECUTE IMMEDIATE '
                                        BEGIN
                                            TRUNCATE TABLE {dest_fqn};
                                            INSERT INTO {dest_fqn}
                                            SELECT {names}
                                            FROM {APP_NAME}.results.{result_table_name}
                                            {'' if out_sample else 'WHERE 1=0'};
                                        END;
                                    ';
                                ELSE
                                    -- Create table based on the SELECT
                                    EXECUTE IMMEDIATE '
                                        {escape_for_f_string(createTableLogic)}
                                    ';
                                END IF;
                            END;
                        """)
                    else:
                        if out_sample:
                            _exec(f"""
                                BEGIN
                                    INSERT INTO {dest_fqn}
                                    SELECT {names}
                                    FROM {APP_NAME}.results.{result_table_name};
                                END;
                            """)
                finally:
                    # Always try to drop the result table, even if the insert/create failed.
                    _exec(f"call {APP_NAME}.api.drop_result_table('{result_table_name}');")

    def _build_projection(self, declared_cols, actual_cols, column_fields, out_sample=None):
        # map physical col -> type
        col_type_map = {col.lower(): dtype for col, dtype in column_fields}
        sample_keys = {k.lower() for k in out_sample[0].as_dict()} if out_sample else set()

        fields = []
        ix = 0

        for name in declared_cols:
            if name not in actual_cols:
                # Declared but not present in results
                fields.append(f"NULL as \"{name}\"")
                continue

            colname = f"col{ix:03}"
            ix += 1

            if colname in sample_keys:
                # Actual column exists in sample
                fields.append(f"{colname} as \"{name}\"")
            else:
                # No sample or missing key → fall back to type
                dtype = col_type_map.get(colname.lower(), "VARCHAR")
                fields.append(f"CAST(NULL AS {dtype}) as \"{name}\"")

        return ", ".join(fields)

    def _construct_configure(self):
        config_dict = {}
        # Only set the IVM flag if there is a value in `config`. Otherwise, let
        # `construct_configure` set the default value.
        ivm_flag = self.config.get('reasoner.rule.incremental_maintenance', None)
        if ivm_flag:
            config_dict['ivm.maintenance_level'] = lqp_ir.Value(value=ivm_flag, meta=None)

        # Set semantics_version from config, defaulting to 0
        semantics_version: str | Any = self.config.get('reasoner.rule.lqp.semantics_version', DEFAULT_LQP_SEMANTICS_VERSION)
        config_dict['semantics_version'] = lqp_ir.Value(value=int(semantics_version), meta=None)

        # Warn if a non-default semantics version is used. Most likely, this is due to a
        # user manually reverting to an older version. We want them to not get stuck on that
        # version for longer than necessary.
        if semantics_version != DEFAULT_LQP_SEMANTICS_VERSION:
            debugging.warn(NonDefaultLQPSemanticsVersionWarning(semantics_version, DEFAULT_LQP_SEMANTICS_VERSION))

        return construct_configure(config_dict, None)

    def _should_sync(self, model) :
        if self._last_model != model:
            return lqp_ir.Sync(fragments=[], meta=None)
        else :
            return None

    def _compile_intrinsics(self) -> lqp_ir.Epoch:
        """Construct an epoch that defines a number of built-in definitions used by the
        emitter."""
        with debugging.span("compile_intrinsics") as span:
            span["compile_type"] = "intrinsics"

            now = self.intrinsic_overrides.get('datetime_now', datetime.now(timezone.utc))

            debug_info = lqp_ir.DebugInfo(id_to_orig_name={}, meta=None)
            intrinsics_fragment = lqp_ir.Fragment(
                id = lqp_ir.FragmentId(id=b"__pyrel_lqp_intrinsics", meta=None),
                declarations = [mk_intrinsic_datetime_now(now)],
                debug_info = debug_info,
                meta = None,
            )

            return lqp_ir.Epoch(
                writes=[
                    lqp_ir.Write(write_type=lqp_ir.Define(fragment=intrinsics_fragment, meta=None), meta=None)
                ],
                meta=None,
            )

    # [RAI-40997] We eagerly undefine query fragments so they are not committed to storage
    def _compile_undefine_query(self, query_epoch: lqp_ir.Epoch) -> lqp_ir.Epoch:
        fragment_ids = []

        for write in query_epoch.writes:
            if isinstance(write.write_type, lqp_ir.Define):
                fragment_ids.append(write.write_type.fragment.id)

        # Construct new Epoch with Undefine operations for all collected fragment IDs
        undefine_writes = [
            lqp_ir.Write(
                write_type=lqp_ir.Undefine(fragment_id=frag_id, meta=None),
                meta=None
            )
            for frag_id in fragment_ids
        ]

        return lqp_ir.Epoch(
            writes=undefine_writes,
            meta=None,
        )

    def compile_lqp(self, model: ir.Model, task: ir.Task):
        configure = self._construct_configure()
        # Merge the epochs into a single transaction. Long term the query bits should all
        # go into a WhatIf action and the intrinsics could be fused with either of them. But
        # for now we just use separate epochs.
        epochs = []
        epochs.append(self._compile_intrinsics())

        sync = self._should_sync(model)

        if sync is not None:
            with debugging.span("compile", metamodel=model) as install_span:
                install_span["compile_type"] = "model"
                _, model_epoch = self.compiler.compile(model, {"fragment_id": b"model"})
                epochs.append(model_epoch)
                self._last_model = model

        with debugging.span("compile", metamodel=task) as txn_span:
            query = f.compute_model(f.logical([task]))
            options = {
                "wide_outputs": self.wide_outputs,
                "fragment_id": b"query",
            }
            result, final_model = self.compiler.compile_inner(query, options)
            export_info, query_epoch = result

            epochs.append(query_epoch)
            epochs.append(self._compile_undefine_query(query_epoch))

            txn_span["compile_type"] = "query"
            txn = mk_transaction(epochs=epochs, configure=configure, sync=sync)
            txn_span["lqp"] = lqp_print.to_string(txn, {"print_names": True, "print_debug": False, "print_csv_filename": False})

        validate_lqp(txn)

        txn_proto = convert_transaction(txn)
        return final_model, export_info, txn_proto

    # TODO (azreika): This should probably be split up into exporting and other processing. There are quite a lot of arguments here...
    def _process_results(self, task: ir.Task, final_model: ir.Model, raw_results: TransactionAsyncResponse, export_info: Optional[tuple], export_to: Optional[Table], update: bool) -> DataFrame:
        cols, extra_cols = self._compute_cols(task, final_model)

        df, errs = result_helpers.format_results(raw_results, cols)
        self.report_errors(errs)

        # Rename columns if wide outputs is enabled
        if self.wide_outputs and len(cols) - len(extra_cols) == len(df.columns):
            df.columns = cols[: len(df.columns)]

        # Process exports
        if export_to and not self.dry_run:
            assert cols, "No columns found in the output"
            assert isinstance(raw_results, TransactionAsyncResponse) and raw_results.transaction, "Invalid transaction result"

            result_cols = export_to._col_names

            if result_cols is not None:
                assert all(col in result_cols or col in extra_cols for col in cols)
            else:
                result_cols = [col for col in cols if col not in extra_cols]
            assert result_cols

            assert export_info, "Export info should be populated if we are exporting results"
            self._export(raw_results.transaction['id'], export_info, export_to, cols, result_cols, update)

        return self._postprocess_df(self.config, df, extra_cols)

    def execute(self, model: ir.Model, task: ir.Task, format: Literal["pandas", "snowpark"] = "pandas",
                export_to: Optional[Table] = None,
                update: bool = False, meta: dict[str, Any] | None = None) -> DataFrame:
        self.prepare_data()
        previous_model = self._last_model

        final_model, export_info, txn_proto = self.compile_lqp(model, task)

        if self.dry_run:
            return DataFrame()

        if format != "pandas":
            raise ValueError(f"Unsupported format: {format}")

        # Format meta as headers
        json_meta = prepare_metadata_for_headers(meta)
        headers = {QUERY_ATTRIBUTES_HEADER: json_meta} if json_meta else {}
        raw_results = self.resources.exec_lqp(
            self.database,
            self.engine,
            txn_proto.SerializeToString(),
            # Current strategy is to run all queries as write transactions, in order to
            # benefit from view caching. This will have to be revisited, because write
            # transactions are serialized.
            readonly=False,
            nowait_durable=True,
            headers=headers,
        )
        assert isinstance(raw_results, TransactionAsyncResponse)

        try:
            return self._process_results(task, final_model, raw_results, export_info, export_to, update)
        except Exception as e:
            # If processing the results failed, revert to the previous model.
            self._last_model = previous_model
            raise e



from __future__ import annotations

from v0.relationalai.semantics.internal import internal as b
from .std import _Number, _make_expr
from math import pi

def abs(value: _Number) -> b.Expression:
    return _make_expr("abs", value, b.Number.ref("res"))

def natural_log(value: _Number) -> b.Expression:
    return _make_expr("natural_log", value, b.Number.ref("res"))

def log(value: _Number, base: _Number | None = None) -> b.Expression:
    if base is None:
        return _make_expr("natural_log", value, b.Number.ref("res"))
    return _make_expr("log", base, value, b.Number.ref("res"))

def log2(value: _Number) -> b.Expression:
    return _make_expr("log", 2, value, b.Number.ref("res"))

def log10(value: _Number) -> b.Expression:
    return _make_expr("log10", value, b.Number.ref("res"))

def sqrt(value: _Number) -> b.Expression:
    return _make_expr("sqrt", value, b.Float.ref("res"))

def maximum(left: _Number, right: _Number) -> b.Expression:
    return _make_expr("maximum", left, right, b.Number.ref("res"))

def minimum(left: _Number, right: _Number) -> b.Expression:
    return _make_expr("minimum", left, right, b.Number.ref("res"))

def isinf(value: _Number) -> b.Expression:
    return _make_expr("isinf", value)

def isnan(value: _Number) -> b.Expression:
    return _make_expr("isnan", value)

def ceil(value: _Number) -> b.Expression:
    return _make_expr("ceil", value, b.Number.ref("res"))

def floor(value: _Number) -> b.Expression:
    return _make_expr("floor", value, b.Number.ref("res"))

def pow(base: _Number, exponent: _Number) -> b.Expression:
    return _make_expr("pow", base, exponent, b.Float.ref("res"))

def cbrt(value: _Number) -> b.Expression:
    return _make_expr("cbrt", value, b.Float.ref("res"))

def factorial(value: _Number) -> b.Expression:
    return _make_expr("factorial", value, b.Number.ref("res"))

def cos(value: _Number) -> b.Expression:
    return _make_expr("cos", value, b.Float.ref("res"))

def cosh(value: _Number) -> b.Expression:
    return _make_expr("cosh", value, b.Float.ref("res"))

def acos(value: _Number) -> b.Expression:
    return _make_expr("acos", value, b.Float.ref("res"))

def acosh(value: _Number) -> b.Expression:
    return _make_expr("acosh", value, b.Float.ref("res"))

def sin(value: _Number) -> b.Expression:
    return _make_expr("sin", value, b.Float.ref("res"))

def sinh(value: _Number) -> b.Expression:
    return _make_expr("sinh", value, b.Float.ref("res"))

def asin(value: _Number) -> b.Expression:
    return _make_expr("asin", value, b.Float.ref("res"))

def asinh(value: _Number) -> b.Expression:
    return _make_expr("asinh", value, b.Float.ref("res"))

def tan(value: _Number) -> b.Expression:
    return _make_expr("tan", value, b.Float.ref("res"))

def tanh(value: _Number) -> b.Expression:
    return _make_expr("tanh", value, b.Float.ref("res"))

def atan(value: _Number) -> b.Expression:
    return _make_expr("atan", value, b.Float.ref("res"))

def atanh(value: _Number) -> b.Expression:
    return _make_expr("atanh", value, b.Float.ref("res"))

def cot(value: _Number) -> b.Expression:
    return _make_expr("cot", value, b.Float.ref("res"))

def acot(value: _Number) -> b.Expression:
    return _make_expr("acot", value, b.Float.ref("res"))

def degrees(value: _Number) -> b.Expression:
    divisor = pi / 180.0
    return _make_expr("/", value, divisor, b.Float.ref("res"))

def radians(value: _Number) -> b.Expression:
    divisor = 180.0 / pi
    return _make_expr("/", value, divisor, b.Float.ref("res"))

def exp(value: _Number) -> b.Expression:
    return _make_expr("exp", value, b.Float.ref("res"))

def erf(value: _Number) -> b.Expression:
    return _make_expr("erf", value, b.Float.ref("res"))

def erfinv(value: _Number) -> b.Expression:
    return _make_expr("erfinv", value, b.Float.ref("res"))

def haversine(x1: _Number, y1: _Number, x2: _Number, y2: _Number, r: _Number) -> b.Expression:
    # 2 * r * asin[sqrt[sin[(x2 - x1)/2] ^ 2 + cos[x1] * cos[x2] * sin[(y2 - y1) / 2] ^ 2]]
    # sin[(x2 - x1)/2] ^ 2
    x_diff = _make_expr("-", x2, x1, b.Float.ref("x_diff"))
    x_diff2 = _make_expr("/", x_diff, 2.0, b.Float.ref("x_diff2"))
    sin_x_diff = _make_expr("sin", x_diff2, b.Float.ref("sin_x_diff"))
    sin_x_pow = _make_expr("pow", sin_x_diff, 2.0, b.Float.ref("sin_x_pow"))

    # cos[x1] * cos[x2]
    cos_x1 = _make_expr("cos", x1, b.Float.ref("cos_x1"))
    cos_x2 = _make_expr("cos", x2, b.Float.ref("cos_x2"))
    cos_x1_x2 = _make_expr("*", cos_x1, cos_x2, b.Float.ref("cos_x1_x2"))

    # sin[(y2 - y1) / 2] ^ 2
    y_diff = _make_expr("-", y2, y1, b.Float.ref("y_diff"))
    y_diff2 = _make_expr("/", y_diff, 2.0, b.Float.ref("y_diff2"))
    sin_y_diff = _make_expr("sin", y_diff2, b.Float.ref("sin_y_diff"))
    sin_y_pow = _make_expr("pow", sin_y_diff, 2.0, b.Float.ref("sin_y_pow"))

    # cos[x1] * cos[x2] * sin[(y2 - y1) / 2] ^ 2
    prod = _make_expr("*", cos_x1_x2, sin_y_pow, b.Float.ref("prod"))
    # sin[(x2 - x1)/2] ^ 2 + cos[x1] * cos[x2] * sin[(y2 - y1) / 2] ^ 2
    prod_sum = _make_expr("+", sin_x_pow, prod, b.Float.ref("prod_sum"))

    sqrt_val = _make_expr("sqrt", prod_sum, b.Float.ref("sqrt_val"))
    asin_val = _make_expr("asin", sqrt_val, b.Float.ref("asin_val"))
    haversine_r = _make_expr("*", r, asin_val, b.Float.ref("haversine_r"))
    haversine_final = _make_expr("*", haversine_r, 2.0, b.Float.ref("haversine_final"))
    return haversine_final



from __future__ import annotations
from typing import Optional

from . import metamodel as mm

class Relations:
    """ Container for built-in relations in a library. """
    def __init__(self):
        self.data = dict[str, mm.Relation]()

    def __getattr__(self, attr: str) -> mm.Relation:
        return self.data[attr]

    def __getitem__(self, attr: str) -> mm.Relation|None:
        return self.data.get(attr)

class CoreLibrary:
    """ Declare some core types and relations as class attributes for performant access. """
    # abstract types
    Any: mm.ScalarType
    AnyEntity: mm.ScalarType
    Numeric: mm.ScalarType
    TypeVar: mm.ScalarType
    EntityTypeVar: mm.ScalarType
    Number: mm.ScalarType
    ScaledNumber: mm.ScalarType
    # meta types
    Field: mm.ScalarType
    Relation: mm.ScalarType
    Type: mm.ScalarType
    Enum: mm.ScalarType
    # concrete primitive types
    Boolean: mm.ScalarType
    String: mm.ScalarType
    Date: mm.ScalarType
    DateTime: mm.ScalarType
    Float: mm.ScalarType
    Hash: mm.ScalarType
    # type aliases
    Integer: mm.NumberType
    DefaultNumber: mm.NumberType
    # binary operators
    plus: mm.Relation
    minus: mm.Relation
    mul: mm.Relation
    div: mm.Relation
    trunc_div: mm.Relation
    power: mm.Relation
    mod: mm.Relation
    # comparison operators
    eq: mm.Relation
    lt: mm.Relation
    lte: mm.Relation
    gt: mm.Relation
    gte: mm.Relation
    neq: mm.Relation
    # typer relations
    cast: mm.Relation

    def __getitem__(self, attr: str) -> mm.Relation:
        if attr in builtins._libraries["core"].data:
            return builtins._libraries["core"].data[attr]
        raise AttributeError(f"No built-in core attribute named '{attr}'.")

    def __iter__(self):
        yield from builtins._libraries["core"].data.values()

class Builtins:
    # Registered libraries by name
    _libraries = dict[str, Relations]()
    # Special case the core library for more performant access
    core = CoreLibrary()

    def register(self, library_name: str, relation: mm.Relation):
        """
            Register a built-in relation under the given library name.
        """
        if library_name not in self._libraries:
            self._libraries[library_name] = Relations()
        self._libraries[library_name].data[relation.name] = relation
        if library_name == "core":
            setattr(CoreLibrary, relation.name, relation)

    # Dynamic access to libraries (e.g., builtins.math.abs)
    def __getattr__(self, attr: str) -> Relations:
        if attr in self._libraries:
            return self._libraries[attr]
        raise AttributeError(f"No built-in library named '{attr}'.")


# Singleton instance of the builtins container
builtins = Builtins()

#------------------------------------------------------
# Helpers
#------------------------------------------------------
def is_function(r: mm.Relation) -> bool:
    """ True if the relation is a function (i.e., it has inputs and has a single output). """
    if len(r.fields) <= 1:
        return False
    # at least 2 fields, only 1 can be output
    output_fields = [f for f in r.fields if not f.input]
    return len(output_fields) == 1

def is_monotyped(r: mm.Relation) -> bool:
    if len(r.fields) == 0:
        return True

    # all types must be the same
    first = r.fields[0].type
    if any(f.type != first for f in r.fields):
        return False

    if r.overloads:
        # every overload must have all fields of the same type
        for overload in r.overloads:
            first = overload.types[0]
            if any(f_type != first for f_type in overload.types):
                return False
    return True

def is_placeholder(relation: mm.Relation) -> bool:
    """ Whether this relation is an unresolved placeholder that needs to be resolved. """
    return isinstance(relation, mm.UnresolvedRelation)

def is_abstract(t: mm.Type) -> bool:
    """ True if the type is abstract, or a collection or union containing an abstract type.
    """
    if isinstance(t, mm.ScalarType):
        return t in [
            builtins.core.Any,
            builtins.core.AnyEntity,
            builtins.core.Number,
            builtins.core.ScaledNumber,
            builtins.core.Numeric,
            builtins.core.TypeVar,
            builtins.core.EntityTypeVar
        ]
    elif isinstance(t, mm.ListType):
        return is_abstract(t.element_type)
    elif isinstance(t, mm.UnionType):
        return any(is_abstract(t) for t in t.types)
    else:
        return False

def is_type_var(t: mm.Type) -> bool:
    return t in [
        builtins.core.TypeVar,
        builtins.core.EntityTypeVar,
    ]

def is_concrete(t: mm.Type) -> bool:
    return not is_abstract(t)

def is_number(t: mm.Type) -> bool:
    """ True if the type is a number type. """
    # note that builtins.core.Number is a ScalarType that represents "any number type",
    # whereas concrete numbers are represented by NumberType instances.
    return isinstance(t, mm.NumberType) or t == builtins.core.Number or t == builtins.core.ScaledNumber

def is_numeric(t: mm.Type) -> bool:
    """ True if the type is numeric (Number or Float). """
    # TODO - we could check by supertype instead, but that would be slower
    return t == builtins.core.Numeric or t == builtins.core.Float or is_number(t)

def is_primitive(t: mm.Type) -> bool:
    """
        True if the type is a primitive type (boolean, string, date, datetime, float, number).
    """
    return t in [
        CoreLibrary.Boolean,
        CoreLibrary.String,
        CoreLibrary.Date,
        CoreLibrary.DateTime,
        CoreLibrary.Float
    ] or is_number(t)

def is_value_type(type: mm.Type):
    """ True if the type extends a primitive type (i.e., is a value type). """
    return isinstance(type, mm.ScalarType) and not is_primitive(type) and get_primitive_supertype(type) is not None

def get_primitive_supertype(type: mm.Type) -> Optional[mm.Type]:
    """ Search the type hierarchy for a primitive supertype. """
    if isinstance(type, mm.ScalarType):
        if is_primitive(type):
            return type
        # walk the hierarchy to find a base primitive
        for parent in type.super_types:
            if found := get_primitive_supertype(parent):
                return found
    if isinstance(type, mm.UnionType):
        for t in type.types:
            if found := get_primitive_supertype(t):
                return found
    return None

def get_number_supertype(type: mm.Type) -> Optional[mm.NumberType]:
    x = get_primitive_supertype(type)
    if isinstance(x, mm.NumberType):
        return x
    return None

def extends(type: mm.Type, supertype: mm.Type) -> bool:
    """ True if `type` extends `supertype` in the type hierarchy. """
    if type == supertype:
        return True
    if isinstance(type, mm.ScalarType):
        for parent in type.super_types:
            if extends(parent, supertype):
                return True
    return False



from abc import abstractmethod

from v0.relationalai.early_access.dsl.core.builders import ExprBuilder
from v0.relationalai.early_access.dsl.core.exprs import Expr, contextStack

# Arithmetic expressions are those Numbers, ScalarVariables, and larger
# expressions that combine these with arithmetic operators
#
class ScalarExpr(Expr):

    # ScalarExpr building operators and methods

    def __add__(self, other):
        return contextStack.root_context().build_plus(self, other)

    def __radd__(self, other):
        return contextStack.root_context().build_plus(self, other)

    def __mod__(self, other):
        return contextStack.root_context().build_modulus(self, other)

    def __rmod__(self, other):
        return contextStack.root_context().build_modulus(self, other)

    def __mul__(self, other):
        return contextStack.root_context().build_times(self, other)

    def __rmul__(self, other):
        return contextStack.root_context().build_times(self, other)

    def __neg__(self):
        return contextStack.root_context().build_negate(self)

    def __sub__(self, other):
        return contextStack.root_context().build_minus(self, other)

    def __rsub__(self, other):
        return contextStack.root_context().build_minus(self, other)

    def __truediv__(self, other):
        return contextStack.root_context().build_divide(self, other)

    def __rtruediv__(self, other):
        return contextStack.root_context().build_divide(self, other)

    # Constraint-building operators and methods

    def __ge__(self, other):
        contextStack.root_context().build_comparison(self, ">=", other)

    def __gt__(self, other):
        contextStack.root_context().build_comparison(self, ">", other)

    def __le__(self, other):
        contextStack.root_context().build_comparison(self, "<=", other)

    def __lt__(self, other):
        contextStack.root_context().build_comparison(self, "<", other)

    def __ne__(self, other): # type: ignore
        contextStack.root_context().build_comparison(self, "!=", other)

    def is_in(self, other):
        return contextStack.root_context().build_element_of(self, other)

    def not_in(self, other):
        return contextStack.root_context().build_not_element_of(self, other)

    @abstractmethod
    def map_builder(self, builder):
        raise NotImplementedError(f"Not implemented for \"{type(self)}\".")


class ScalarVariable(ScalarExpr):

    def physical_typeof(self):
        raise Exception(f"Cannot retrieve phsyical type of variable {self.display()}")

    def grounded(self):
        return False

    def grounded_using(self, groundings):
        return self.display() in groundings

    def refersto(self, varname: str) -> bool:
        return self.display() == varname

    def scalar_refs(self):
        return {self.display(): self}

    def rename(self, renaming):
        vname = self.display()
        if vname in renaming:
            return renaming[vname]
        else:
            return self

    def substitute(self, bindings):
        return self.rename(bindings)

    def typeof(self):
        pass

    def variable(self):
        return True

class Literal(ScalarExpr):
    def __init__(self, v): self.val = v

    def display(self): return str(self.val)

    def entityid(self): return hash(self.val)

    def grounded(self): return True

    def grounded_using(self, groundings): return True

    def rename(self, renaming): return self

    def revar(self, vmap): return self

    def refersto(self, varname): return False

    def simplify(self): return self

    def substitute(self, bindings): return self

    def __hash__(self):
        return hash(self.val)

    def map_builder(self, builder):
        return self

class Number(Literal):
    pass

class String(Literal):

    def display(self): return f"\"{self.val}\""


class Negate(ScalarExpr):
    def __init__(self, x):
        self._part = x

    def display(self):
        p = self._part
        return f"- {p.display()}"

    def entityid(self):
        return hash((Negate, self._part.entityid()))

    # Attempts to evaluate this expression statically, returning a single Number if
    # possible, otherwise returns self.
    #
    def evaluate(self):
        if self.grounded():
            exec(f"self.x = {self.display()}")
            return Number(self.x) # type: ignore
        else:
            return self

    def grounded(self):
        return self._part.grounded()

    def grounded_using(self, groundings):
        return self._part.grounded_using(groundings)

    def scalar_refs(self):
        return self._part.scalar_refs()

    def refersto(self, varname):
        return self._part.refersto(varname)

    def rename(self, renaming):
        return Negate(self._part.rename(renaming))

    def revar(self, vmap):
        return Negate(self._part.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Negate(self._part.simplify())

    def substitute(self, bindings):
        return Negate(self._part.substitute(bindings))


class BinaryScalarExpr(ScalarExpr):

    def __init__(self, op):
        self._op = op

    def display(self):
        return f"{self._left.display()} {self._op} {self._right.display()}"

    # Attempts to evaluate this expression statically, returning a single Number if
    # possible, otherwise returns self.
    #
    def evaluate(self):
        if self.grounded():
            exec(f"self.x = {self.display()}")
            return Number(self.x) # type: ignore
        else:
            return self

    def grounded(self):
        return self._left.grounded() and self._right.grounded()

    def grounded_using(self, groundings):
        return self._left.grounded_using(groundings) and self._right.grounded_using(groundings)

    def scalar_refs(self):
        dic = self._left.scalar_refs()
        rdict = self._right.scalar_refs()
        for v in rdict:
            if v not in dic:
                dic[v] = rdict[v]
        return dic

    def refersto(self, varname):
        return self._left.refersto(varname) or self._right.refersto(varname)

    def set_args(self, x, z):
        self._left = box_number(x)
        self._right = box_number(z)

    def op(self):
        return self._op


class Plus(BinaryScalarExpr):

    def __init__(self, x, z):
        super().__init__("+")
        self.set_args(x, z)

    def rename(self, renaming):
        return Plus(self._left.rename(renaming),
                    self._right.rename(renaming))

    def revar(self, vmap):
        return Plus(self._left.revar(vmap),
                    self._right.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Plus(self._left.simplify(), self._right.simplify())

    def substitute(self, bindings):
        return Plus(self._left.substitute(bindings), self._right.substitute(bindings))

    def entityid(self):
        return hash((Plus,
                     self._left.entityid(),
                     self._right.entityid()))

    def map_builder(self, builder):
        return Plus(self._left.map_builder(builder), self._right.map_builder(builder))


class Minus(BinaryScalarExpr):

    def __init__(self, x, z):
        super().__init__("-")
        self.set_args(x, z)

    def rename(self, renaming):
        return Minus(self._left.rename(renaming),
                     self._right.rename(renaming))

    def revar(self, vmap):
        return Minus(self._left.revar(vmap),
                     self._right.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Minus(self._left.simplify(), self._right.simplify())

    def substitute(self, bindings):
        return Minus(self._left.substitute(bindings), self._right.substitute(bindings))

    def entityid(self):
        return hash((Minus,
                     self._left.entityid(),
                     self._right.entityid()))

    def map_builder(self, builder):
        return Minus(self._left.map_builder(builder), self._right.map_builder(builder))


class Modulus(BinaryScalarExpr):
    def __init__(self, x, z):
        super().__init__("%")
        self.set_args(x, z)

    def rename(self, renaming):
        return Modulus(self._left.rename(renaming),
                       self._right.rename(renaming))

    def revar(self, vmap):
        return Modulus(self._left.revar(vmap),
                       self._right.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Modulus(self._left.simplify(), self._right.simplify())

    def substitute(self, bindings):
        return Modulus(self._left.substitute(bindings), self._right.substitute(bindings))

    def entityid(self):
        return hash((Modulus,
                     self._left.entityid(),
                     self._right.entityid()))


class Times(BinaryScalarExpr):

    def __init__(self, x, z):
        super().__init__("*")
        self.set_args(x, z)

    def rename(self, renaming):
        return Times(self._left.rename(renaming),
                     self._right.rename(renaming))

    def revar(self, vmap):
        return Times(self._left.revar(vmap),
                     self._right.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Times(self._left.simplify(), self._right.simplify())

    def substitute(self, bindings):
        return Times(self._left.substitute(bindings), self._right.substitute(bindings))

    def entityid(self):
        return hash((Times,
                     self._left.entityid(),
                     self._right.entityid()))


class Divide(BinaryScalarExpr):

    def __init__(self, x, z):
        super().__init__("/")
        self.set_args(x, z)

    def entityid(self):
        return hash((Divide,
                     self._left.entityid(),
                     self._right.entityid()))

    def rename(self, renaming):
        return Divide(self._left.rename(renaming),
                      self._right.rename(renaming))

    def revar(self, vmap):
        return Divide(self._left.revar(vmap),
                      self._right.revar(vmap))

    def simplify(self):
        ev = self.evaluate()
        if ev.entityid() != self.entityid():
            return ev
        else:
            return Divide(self._left.simplify(), self._right.simplify())

    def substitute(self, bindings):
        return Divide(self._left.substitute(bindings), self._right.substitute(bindings))


class ScalarExprBuilder(ExprBuilder):

    def build_divide(self, left, right): return Divide(left, box_number(right))

    def build_plus(self, left, right): return Plus(left, box_number(right))

    def build_minus(self, left, right): return Minus(left, box_number(right))

    def build_modulus(self, left, right): return Modulus(left, box_number(right))

    def build_negate(self, expr): return Negate(expr)

    def build_times(self, left, right): return Times(left, box_number(right))


def box_number(x):
    if isinstance(x, (int, float)):
        return Number(x)
    else:
        if isinstance(x, str):
            return String(x)
        else:
            return x



from itertools import count
from v0.relationalai.semantics.metamodel import ir, types, util
from v0.relationalai.semantics.internal import internal as b
from v0.relationalai.semantics.snowflake import Table

def reset_state():
    """
    Reset global state for consistent test snapshots.

    When we execute a pyrel program we accumulate some state, such as custom decimals, and
    we increase the counter for object ids. This function resets those counters and other
    state to a known baseline to ensure that test snapshots are consistent if we run the
    test alone vs in a suite of tests.
    """
    # reset the global id counters
    ir._global_id = count(10000)
    b._global_id = count(10000)
    b._global_roots = util.ordered_set()

    # Used to generate error source IDs
    b.errors.ModelError.error_locations.clear()

    # reset the ErrorConcept state, so that Error.new() always generates the same IR
    b.ErrorConcept._error_props.clear()
    b.ErrorConcept._relation = None
    b.ErrorConcept._overloads.clear()

    # cleanup relationships declared on global concepts
    concepts_to_cleaanup = [b.Concept.builtins["Error"], b.Concept.builtins["Any"]]
    keep = ["shape", "builtins", "globals"]
    for concept in concepts_to_cleaanup:
        for attr in list(concept._relationships.keys()):
            if attr not in keep:
                del concept._relationships[attr]
        concept.globals.clear()

    # caches of custom decimals
    for k in list(types._decimal_types.keys()):
        if types._decimal_types[k] != types.Decimal:
            del types._decimal_types[k]
    for k in list(b.Concept.builtins):
        if k.startswith("Decimal(") and b.Concept.builtins[k] is not b.Decimal:
            del b.Concept.builtins[k]

    # clear any cached Table sources
    Table._used_sources.clear()


import warnings

from v0.relationalai.semantics.metamodel.helpers import collect_implicit_vars

__all__ = ["collect_implicit_vars"]

warnings.warn(
    "relationalai.early_access.metamodel.helpers is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.helpers",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations

from . import NumberValue, FloatValue, _get_number_value
from ..frontend.base import Library, Expression, Field
from ..frontend.core import Numeric, Number, Integer, Float, make_overloads, minus, mul, div, power

# the front-end library object
library = Library("math")

# ------------------------------
# Basics
# ------------------------------
_abs = library.Relation("abs", [Field.input("value", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 2))
_ceil = library.Relation("ceil", [Field.input("value", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 2))
_clip = library.Relation("clip", [Field.input("value", Numeric), Field.input("lower", Numeric), Field.input("upper", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 4))
_factorial = library.Relation("factorial", [Field.input("value", Integer), Field("result", Integer)])
_floor = library.Relation("floor", [Field.input("value", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 2))
_isclose = library.Relation("isclose", [Field.input("tolerance", Numeric), Field.input("x", Numeric), Field.input("y", Numeric)], overloads=make_overloads([Number, Float], 3))
_sign = library.Relation("sign", [Field.input("x", Numeric), Field("result", Integer)], overloads=[[Number, Integer], [Float, Integer]])
_trunc_divide = library.Relation("trunc_divide", [Field.input("numerator", Number), Field.input("denominator", Number), Field("result", Number)])
_maximum = library.Relation("maximum", [Field.input("left", Numeric), Field.input("right", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 3))
_minimum = library.Relation("minimum", [Field.input("left", Numeric), Field.input("right", Numeric), Field("result", Numeric)], overloads=make_overloads([Number, Float], 3))
_isinf = library.Relation("isinf", [Field.input("value", Float)])
_isnan = library.Relation("isnan", [Field.input("value", Float)])

def abs(value: NumberValue) -> Expression:
    return _abs(value)

def ceil(value: NumberValue) -> Expression:
    return _ceil(value)

def clip(value: NumberValue, lower: NumberValue, upper: NumberValue) -> Expression:
    # CLIP = CLAMP(v,min,max) = LEAST(max,GREATEST(min,v))
    return _clip(lower, upper, value)

def factorial(value: NumberValue) -> Expression:
    v = _get_number_value(value)
    if v is not None and v < 0:
        raise ValueError("Cannot take the factorial of a negative number")
    return _factorial(value)

def floor(value: NumberValue) -> Expression:
    return _floor(value)

def isclose(x: NumberValue, y: NumberValue, tolerance: NumberValue = 1e-9) -> Expression:
    # APPROX_EQUAL = ABS(x - y) < tolerance
    return _isclose(tolerance, x, y)

def sign(x: NumberValue) -> Expression:
    return _sign(x)

def trunc_divide(numerator: NumberValue, denominator: NumberValue) -> Expression:
    return _trunc_divide(numerator, denominator)

def maximum(left: NumberValue, right: NumberValue) -> Expression:
    # GREATEST
    return _maximum(left, right)

def minimum(left: NumberValue, right: NumberValue) -> Expression:
    # LEAST
    return _minimum(left, right)

def isinf(value: FloatValue) -> Expression:
    return _isinf(value)

def isnan(value: FloatValue) -> Expression:
    return _isnan(value)

# -------------------------------
# Power and Logarithmic Functions
# -------------------------------

_cbrt = library.Relation("cbrt", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_sqrt = library.Relation("sqrt", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_exp = library.Relation("exp", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_power = library.Relation("power", [Field.input("base", Numeric), Field.input("exponent", Numeric), Field("result", Float)],
                          overloads=[
                              #[Number, Number, Float], # v0 and engines are not handling this correctly
                              [Float, Float, Float]])
_log = library.Relation("log", [Field.input("base", Numeric), Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Number, Float], [Float, Float, Float]])
_natural_log = library.Relation("natural_log", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])

def cbrt(value: NumberValue) -> Expression:
    return _cbrt(value)

def sqrt(value: NumberValue) -> Expression:
    v = _get_number_value(value)
    if v is not None and v < 0:
        raise ValueError("Cannot take the square root of a negative number")
    return _sqrt(value)

def exp(value: NumberValue) -> Expression:
    return _exp(value)

def pow(base: NumberValue, exponent: NumberValue) -> Expression:
    return _power(base, exponent)

def natural_log(value: NumberValue) -> Expression:
    return _natural_log(value)

def log(x: NumberValue, base: NumberValue | None = None) -> Expression:
    v = _get_number_value(x)
    if v is not None and v <= 0:
        raise ValueError("Cannot take the logarithm of a non-positive number")
    if base is None:
        return natural_log(x)
    return _log(base, x)

def log2(value: NumberValue) -> Expression:
    return log(value, 2)

def log10(value: NumberValue) -> Expression:
    return log(value, 10)



# ------------------------------
# Trigonometry
# ------------------------------

_degrees = library.Relation("degrees", [Field.input("radians", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_radians = library.Relation("radians", [Field.input("degrees", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_cos = library.Relation("cos", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_sin = library.Relation("sin", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_tan = library.Relation("tan", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_cot = library.Relation("cot", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_acos = library.Relation("acos", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_asin = library.Relation("asin", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_atan = library.Relation("atan", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_acot = library.Relation("acot", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_cosh = library.Relation("cosh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_sinh = library.Relation("sinh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_tanh = library.Relation("tanh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_acosh = library.Relation("acosh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_asinh = library.Relation("asinh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_atanh = library.Relation("atanh", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])

def degrees(radians: NumberValue) -> Expression:
    return _degrees(radians)

def radians(degrees: NumberValue) -> Expression:
    return _radians(degrees)

def cos(value: NumberValue) -> Expression:
    return _cos(value)

def sin(value: NumberValue) -> Expression:
    return _sin(value)

def tan(value: NumberValue) -> Expression:
    return _tan(value)

def cot(value: NumberValue) -> Expression:
    return _cot(value)

def acos(value: NumberValue) -> Expression:
    return _acos(value)

def asin(value: NumberValue) -> Expression:
    return _asin(value)

def atan(value: NumberValue) -> Expression:
    return _atan(value)

def acot(value: NumberValue) -> Expression:
    return _acot(value)

def cosh(value: NumberValue) -> Expression:
    return _cosh(value)

def sinh(value: NumberValue) -> Expression:
    return _sinh(value)

def tanh(value: NumberValue) -> Expression:
    return _tanh(value)

def acosh(value: NumberValue) -> Expression:
    v = _get_number_value(value)
    if v is not None and v < 1:
        raise ValueError("acosh expects a value greater than or equal to 1.")
    return _acosh(value)

def asinh(value: NumberValue) -> Expression:
    return _asinh(value)

def atanh(value: NumberValue) -> Expression:
    v = _get_number_value(value)
    if v is not None and (v <= -1 or v >= 1):
        raise ValueError("atanh expects a value between -1 and 1, exclusive.")
    return _atanh(value)

# ------------------------------
# Special Functions
# ------------------------------
_erf = library.Relation("erf", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])
_erfinv = library.Relation("erfinv", [Field.input("value", Numeric), Field("result", Float)], overloads=[[Number, Float], [Float, Float]])

def erf(value: NumberValue) -> Expression:
    return _erf(value)

def erfinv(value: NumberValue) -> Expression:
    v = _get_number_value(value)
    if v is not None and (v < -1 or v > 1):
        raise ValueError("erfinv expects a value between -1 and 1, inclusive.")
    return _erfinv(value)

def haversine(lat1: NumberValue, lon1: NumberValue, lat2: NumberValue, lon2: NumberValue, radius: NumberValue = 1.0) -> Expression:
    # 2 * r * asin[sqrt[sin[(lat2 - lat1)/2] ^ 2 + cos[lat1] * cos[lat2] * sin[(lon2 - lon1) / 2] ^ 2]]
    return mul(2, radius, _asin(_sqrt(
        power(_sin(div(minus(lat2, lat1), 2)), 2) +
        mul(_cos(lat1), _cos(lat2), power(_sin(div(minus(lon2, lon1), 2)), 2))
    )))



from v0.relationalai.dsl import Symbol, InlineRelation
from v0.relationalai.debugging import logger


#
# This is the schema of the `Config` relation required by the Pathfinder Rel library, in
# versions 0.7+.
#
CONFIG_SCHEMA = {
    "graph_type": {
        "type": "symbol",
        "enum": ["labeled"],
        "default": "labeled"
    },
    "semantics": {
        "type": "symbol",
        "enum": ["shortest_paths"],
        "default": "shortest_paths"
    },
    "group": {
        "type": "symbol",
        "enum": ["any_pair"],
        "default": "any_pair"
    },
    "selector": {
        "type": "symbol",
        "enum": ["all", "single", "limit"],
        "default": "all"
    },
    "path_count": {
        "type": int,
        "default": None
    },
    "max_path_length": {
        "type": int,
        "default": None
    },
    "random_seed": {
        "type": int,
        "default": None
    },
    "path_ids": {
        "type": "symbol",
        "enum": ["canonical", "non_canonical"],
        "default": "non_canonical"
    },
    "node_indexing": {
        "type": "symbol",
        "enum": ["zero_based", "one_based"],
        "default": "zero_based"
    },
    "edge_indexing": {
        "type": "symbol",
        "enum": ["zero_based", "one_based"],
        "default": "one_based"
    },
    "product_graph": {
        "type": "symbol",
        "enum": ["materialized", "non_materialized"],
        "default": "non_materialized"
    },
    "search_strategy": {
        "type": "symbol",
        "enum": ["from_both_sides", "from_source", "from_target"],
        "default": "from_both_sides"
    },
    "path_enum_partials": {
        "type": "symbol",
        "enum": ["dynamic", "static"],
        "default": "dynamic"
    },
    "search_start_nodes": {
        "type": "symbol",
        "enum": ["multiple", "unique"],
        "default": "multiple"
    },
    "search_radius": {
        "type": "symbol",
        "enum": ["unbounded", "bounded"],
        "default": "bounded"
    }
}

DEFAULT_CONFIG = {
    "graph_type": "labeled",
    "semantics": "shortest_paths",
    "group": "any_pair",
    "selector": "all",
    "path_ids": "non_canonical",
    "node_indexing": "zero_based",
    "edge_indexing": "one_based",
    "debug": "debug_off",
    "product_graph": "materialized",
    "search_strategy": "from_both_sides",
    "path_enum_partials": "dynamic",
    "search_start_nodes": "multiple",
    "search_radius": "bounded"
}

#
# Creates an inline relation with the content specified by the `contents` dict
# interpreted with the help of the `schema`.
#
def make_inline_relation(model, contents, schema):
    pairs = []
    for k, v in contents.items():
        if k not in schema:
            logger.info(f"Skipping unknown key {k} with value {v}")
            continue
        elif v is None:
            logger.info(f"Skipping key {k} with null value")
            continue
        elif schema[k]['type'] == 'symbol':
            assert isinstance(v, str), f"Expected a string for {k}, got {v} of type {type(v)}"
            pairs.append((Symbol(k), Symbol(v)))
        else:
            pairs.append((Symbol(k), v))


    logger.info(f"Creating inline relation with {len(pairs)} pairs:")
    for k, v in pairs:
        logger.info(f"  {k._var.value}: {v._var.value if isinstance(v, Symbol) else v}")

    return InlineRelation(model, pairs) # TAG: #pyrel

#
# Uses option values to creates an inlined `Config` relation with relevant configuration
# values for a call of `find_paths` in Pathfinder Rel library.
#
def make_config_relation(model, options):
    config = DEFAULT_CONFIG.copy()
    config.update(options)

    return make_inline_relation(model, config, CONFIG_SCHEMA)

INPUT_SCHEMA = {
    'pg_graph' :{
        'type': 'relation'
    },
    'pg_source': {
        'type': 'relation'
    },
    'pg_target': {
        'type': 'relation'
    }
}

#
# Creates an inline relation with `Input` relation specifying the path query in the form of
# a product graph.
#
def make_input_relation(model, pg_graph, pg_source, pg_target):

    d = {
        'pg_graph': pg_graph,
        'pg_source': pg_source,
        'pg_target': pg_target
    }

    return make_inline_relation(model, d, INPUT_SCHEMA)



from __future__ import annotations
from ..frontend.base import Field, Library, Relationship, Aggregate, Group, TupleVariable, Value, Distinct
from ..frontend.core import Any, Boolean, Number, String, Integer, TypeVar, ScaledNumber, Numeric, Float

library = Library("aggregates")

AggValue = Value | Distinct

#------------------------------------------------------
# Aggregates
#------------------------------------------------------

_sum = library.Relation("sum", fields=[Field.input("value", Numeric), Field("result", Numeric)],
                        overloads=[[Number, Number], [Float, Float]])
_count = library.Relation("count", fields=[Field("result", Integer)])
_avg = library.Relation("avg", fields=[Field.input("over", Numeric), Field("result", Numeric)],
                        overloads=[[Number, ScaledNumber], [Float, Float]])
_min = library.Relation("min", fields=[Field.input("over", TypeVar), Field("result", TypeVar)])
_max = library.Relation("max", fields=[Field.input("over", TypeVar), Field("result", TypeVar)])
_string_join = library.Relation("string_join",
        fields=[Field.input("index", Number), Field.input("sep", String), Field.input("over", String), Field("result", String)])

_sort = library.Relation("sort", fields=[Field.input("limit", Integer), Field.input("args", Any, is_list=True), Field.input("is_asc", Boolean, is_list=True)])
_rank = library.Relation("rank", fields=[Field.input("args", Any, is_list=True), Field.input("is_asc", Boolean, is_list=True), Field("rank", Integer)])
_limit = library.Relation("limit", fields=[Field.input("limit", Integer), Field.input("args", Any, is_list=True), Field.input("is_asc", Boolean, is_list=True)])


def sum(*args: AggValue) -> Aggregate:
    return Aggregate(_sum, *args)

def count(*args: AggValue) -> Aggregate:
    return Aggregate(_count, *args)

def min(*args: AggValue) -> Aggregate:
    return Aggregate(_min, *args)

def max(*args: AggValue) -> Aggregate:
    return Aggregate(_max, *args)

def avg(*args: AggValue) -> Aggregate:
    return Aggregate(_avg, *args)

def string_join(*args: AggValue, sep="", index=1) -> Aggregate:
    return Aggregate(_string_join, index, sep, *args)

class Ordering:
    def __init__(self, *values:AggValue, is_asc=True) -> None:
        self._values = values
        self._is_asc = is_asc

    @staticmethod
    def handle_arg(arg: AggValue|Ordering, ordering, ordering_args, is_asc=True) -> bool:
        has_distinct = False
        if isinstance(arg, Ordering):
            for v in arg._values:
                has_distinct = Ordering.handle_arg(v, ordering, ordering_args, arg._is_asc) or has_distinct
        elif isinstance(arg, Distinct):
            has_distinct = True
            for v in arg._items:
                Ordering.handle_arg(v, ordering, ordering_args, is_asc)
        else:
            ordering.append(is_asc)
            ordering_args.append(arg)
        return has_distinct

    @staticmethod
    def get_ordering_args(args: tuple[AggValue|Ordering, ...]) -> tuple[tuple[AggValue,...], tuple[bool,...], bool]:
        ordering = []
        ordering_args = []
        has_distinct = False
        for arg in args:
            has_distinct = Ordering.handle_arg(arg, ordering, ordering_args) or has_distinct
        return tuple(ordering_args), tuple(ordering), has_distinct

def asc(*args: AggValue) -> Ordering:
    return Ordering(*args, is_asc=True)

def desc(*args: AggValue) -> Ordering:
    return Ordering(*args, is_asc=False)

def _sort_agg(limit: int, *args: AggValue|Ordering) -> Aggregate:
    ordering_args, is_asc, has_distinct = Ordering.get_ordering_args(args)
    return Aggregate(_sort, limit, TupleVariable(ordering_args), TupleVariable(is_asc), distinct=has_distinct)

def rank(*args: AggValue|Ordering) -> Aggregate:
    ordering_args, is_asc, has_distinct = Ordering.get_ordering_args(args)
    return Aggregate(_rank, TupleVariable(ordering_args), TupleVariable(is_asc), distinct=has_distinct)

def limit(limit: int, *args: AggValue|Ordering) -> Aggregate:
    ordering_args, is_asc, has_distinct = Ordering.get_ordering_args(args)
    return Aggregate(_limit, limit, TupleVariable(ordering_args), TupleVariable(is_asc), distinct=has_distinct)

def rank_asc(*args: AggValue) -> Aggregate:
    return Aggregate(_rank, TupleVariable(args), TupleVariable([True for _ in args]))

def rank_desc(*args: AggValue) -> Aggregate:
    return Aggregate(_rank, TupleVariable(args), TupleVariable([False for _ in args]))

def top(limit: int, *args: AggValue) -> Aggregate:
    return Aggregate(_limit, limit, TupleVariable(args), TupleVariable([False for _ in args]))

def bottom(limit: int, *args: AggValue) -> Aggregate:
    return Aggregate(_limit, limit, TupleVariable(args), TupleVariable([True for _ in args]))

#------------------------------------------------------
# Per
#------------------------------------------------------

class Per(Group):

    def sum(self, *args: AggValue) -> Aggregate:
        return sum(*args).per(*self._args)

    def count(self, *args: AggValue) -> Aggregate:
        return count(*args).per(*self._args)

    def min(self, *args: AggValue) -> Aggregate:
        return min(*args).per(*self._args)

    def max(self, *args: AggValue) -> Aggregate:
        return max(*args).per(*self._args)

    def avg(self, *args: AggValue) -> Aggregate:
        return avg(*args).per(*self._args)

    def string_join(self, *args: AggValue, sep="", index=1) -> Aggregate:
        return string_join(*args, sep=sep, index=index).per(*self._args)

    def rank(self, *args: AggValue|Ordering) -> Aggregate:
        return rank(*args).per(*self._args)

    def limit(self, limit_: int, *args: AggValue|Ordering) -> Aggregate:
        return limit(limit_, *args).per(*self._args)

    def rank_asc(self, *args: AggValue) -> Aggregate:
        return rank_asc(*args).per(*self._args)

    def rank_desc(self, *args: AggValue) -> Aggregate:
        return rank_desc(*args).per(*self._args)

    def top(self, limit: int, *args: AggValue) -> Aggregate:
        return top(limit, *args).per(*self._args)

    def bottom(self, limit: int, *args: AggValue) -> Aggregate:
        return bottom(limit, *args).per(*self._args)

def per(*args: Value) -> Per:
    return Per(*args)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Optional, Iterable, Sequence as PySequence, Union

from v0.relationalai import debugging
from v0.relationalai.semantics.metamodel import ir, visitor
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set, group_by



@dataclass(frozen=True)
class Compiler():
    """ Compilers can rewrite a model into a different model, and can compile a model into
    a String, usually of a different language like Rel or SQL. """
    # configurable sequence of passes
    passes: list[Pass]

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        """ Apply a sequence of transformation passes over a model, creating a new model. """
        with debugging.span("passes") as span:
            for p in self.passes:
                with debugging.span(p.name) as span:
                    model = p.rewrite(model, options)
                    span["metamodel"] = str(model.root)
                p.reset()
        return model

    def compile(self, model: ir.Model, options:dict={}) -> Any:
        return self.compile_inner(model, options)[0]

    def compile_inner(self, model: ir.Model, options:dict={}) -> tuple[Any, ir.Model]:
        model = self.rewrite(model, options)
        return self.do_compile(model, options), model

    def do_compile(self, model: ir.Model, options:dict={}) -> Any:
        """ Perform the compilation from model to string. The default implementation simply
        pretty prints the model IR. """
        return ir.node_to_string(model)


@dataclass
class Pass():
    """
    A compiler rewrite pass.
    """
    @property
    def name(self):
        return self.__class__.__name__

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        raise NotImplementedError(f"{self.name} rewrite not implemented")

    def reset(self):
        pass

# TODO - maybe extract these into a compiler_utils module?

def group_tasks(tasks: PySequence[ir.Task], categories: dict[str, Union[type, tuple[type, ...]]]) -> dict[str, OrderedSet[ir.Task]]:
    """ Categorize these tasks in these categories, based on their types. The categories
    dict contains a category name to the set of types that should be in this category. The
    result is a dict from that name to the Tasks in that category. There is a special
    "other" category for uncategorized tasks. """
    def categorize(n):
        for category, types in categories.items():
            if isinstance(n, types):
                return category
        return "other"

    groups = group_by(tasks, lambda t: categorize(t))
    # ensure all categories have an entry
    for key in categories:
        if key not in groups:
            groups[key] = ordered_set()
    if "other" not in groups:
            groups["other"] = ordered_set()
    return groups


@dataclass(frozen=True)
class VarMap:
    """ Maintain a map from variables in the original metamodel to the corresponding
    variables in the rewritten metamodel. """
    mapping: dict[ir.Var, ir.Var] = field(default_factory=dict)

    def get(self, v: ir.Var) -> ir.Var:
        if v not in self.mapping:
            self.mapping[v] = ir.Var(v.type, v.name)
        return self.mapping[v]

    def get_many(self, vars: Iterable[ir.Var]) -> list[ir.Var]:
        return [self.get(v) for v in vars]

    def __contains__(self, var: ir.Var):
        return var in self.mapping


@dataclass
class ReplaceVars(visitor.Rewriter):
    """ A pass that replaces variables found with the mapping from the VarMap. """
    varmap: VarMap = field(init=True)

    def handle_var(self, node: ir.Var, parent: ir.Node, ctx:Optional[Any]=None):
        if node in self.varmap:
            return self.varmap.get(node)
        return node

@dataclass
class DumpModel(Pass):
    """ A pass that dumps the model, verbosely. """
    header: str = field(init=True, default="")

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        if self.header:
            print("-" * 80)
            print(self.header)
            print()
        ir.dump(model)
        print()
        return model

@dataclass
class PrintModel(Pass):
    """ A pass that prints the model. """
    header: str = field(init=True, default="")

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        if self.header:
            print("-" * 80)
            print(self.header)
            print()
        print(model)
        print()
        return model



#!/usr/bin/env python
"""
Snapshot Viewer Tool

This tool generates an HTML report that displays test code and its corresponding
snapshots side by side, making it easier to review snapshot tests.

Usage:
  python tools/snapshot_viewer.py [--output report.html] [--test-dir tests/unit/]
"""

import argparse
import os
import tempfile
import webbrowser
import subprocess
from dataclasses import dataclass
from typing import List, Optional, cast
import json

@dataclass
class TestInfo:
    """Information about a test function."""
    file_path: str
    function_name: str
    class_name: Optional[str]
    code: str
    line_number: int
    snapshot_names: List[str]

@dataclass
class SnapshotInfo:
    """Information about a snapshot file."""
    file_path: str
    content: str
    original_content: Optional[str] = None
    test_info: Optional[TestInfo] = None
    is_modified: bool = False

def generate_html_report(tests, snapshots, output_path):
    """Generate an HTML report of the tests and their snapshots."""
    print("Generating HTML report...")

    # Create HTML content
    html = """
    <!DOCTYPE html>
    <html>
    <head>
        <title>Snapshot Test Report</title>
        <style>
            body {
                font-family: Arial, sans-serif;
                margin: 0;
                padding: 0;
                background-color: #f5f5f5;
            }
            .container {
                max-width: 1200px;
                margin: 0 auto;
                padding: 20px;
            }
            .header {
                background-color: #333;
                color: white;
                padding: 10px 20px;
                margin-bottom: 20px;
                position: sticky;
                top: 0;
                z-index: 100;
            }
            .test-container {
                background-color: white;
                border-radius: 5px;
                box-shadow: 0 2px 5px rgba(0,0,0,0.1);
                margin-bottom: 20px;
                overflow: hidden;
            }
            .test-header {
                background-color: #f0f0f0;
                padding: 10px 15px;
                border-bottom: 1px solid #ddd;
                font-weight: bold;
            }
            .test-content {
                display: flex;
                flex-wrap: wrap;
            }
            .test-code {
                flex: 1;
                min-width: 300px;
                padding: 15px;
                border-right: 1px solid #eee;
                white-space: pre-wrap;
                font-family: monospace;
                overflow-x: auto;
            }
            .snapshot-container {
                flex: 1;
                min-width: 300px;
                padding: 15px;
            }
            .snapshot {
                margin-bottom: 20px;
                border: 1px solid #eee;
                border-radius: 3px;
                overflow: hidden;
            }
            .snapshot-header {
                background-color: #f9f9f9;
                padding: 8px 12px;
                border-bottom: 1px solid #eee;
                font-size: 14px;
                font-weight: bold;
                display: flex;
                flex-direction: row;
                justify-content: space-between;
                align-items: center;
            }
            .snapshot-content {
                padding: 12px;
                white-space: pre-wrap;
                font-family: monospace;
                overflow-x: auto;
                max-height: 500px;
                overflow-y: auto;
            }
            .modified {
                background-color: #fff8e1;
            }
            .no-snapshots {
                padding: 15px;
                color: #999;
                font-style: italic;
            }
            .search-container {
                margin-bottom: 20px;
            }
            #search {
                width: 100%;
                padding: 10px;
                border: 1px solid #ddd;
                border-radius: 4px;
                font-size: 16px;
            }
            .nav-button {
                background-color: #4CAF50;
                border: none;
                color: white;
                padding: 10px 15px;
                text-align: center;
                text-decoration: none;
                display: inline-block;
                font-size: 16px;
                margin: 4px 2px;
                cursor: pointer;
                border-radius: 4px;
            }
            .nav-button:hover {
                background-color: #45a049;
            }
            .nav-button:disabled {
                background-color: #cccccc;
                cursor: not-allowed;
            }
            .fixed-nav {
                position: fixed;
                bottom: 20px;
                right: 20px;
                display: flex;
                gap: 10px;
                z-index: 100;
            }
            .toggle-version-btn {
                padding: 5px 10px;
                background-color: #555;
                color: white;
                border: none;
                border-radius: 4px;
                cursor: pointer;
                font-size: 0.9em;
            }
            .toggle-version-btn:hover {
                background-color: #333;
            }
            .toggle-version-btn.active {
                background-color: #4CAF50;
            }
            .original-content {
                display: none;
            }
            .button-pair {
                display: flex;
                gap: 10px;
            }
        </style>
    </head>
    <body>
        <div class="container">
            <h1>Snapshot Test Report</h1>

            <div class="search-container">
                <input type="text" id="search" placeholder="Search for tests or snapshots...">
            </div>
            <div id="tests">
    """

    # Sort tests by file path and function name
    sorted_tests = sorted(tests, key=lambda t: (t.file_path, t.function_name))

    # Add each test to the HTML
    for test in sorted_tests:
        # Find snapshots for this test
        test_snapshots = []
        for path, info in snapshots.items():
            if info.test_info and info.test_info.file_path == test.file_path and info.test_info.function_name == test.function_name:
                test_snapshots.append(info)

        # Create test container
        html += f"""
        <div class="test-container" data-test="{test.function_name}" data-file="{os.path.basename(test.file_path)}">
            <div class="test-header">
                {"".join(test.file_path.split("/tests/")[1:])} - {test.function_name}
            </div>
            <div class="test-content">
                <div class="test-code">{test.code}</div>
                <div class="snapshot-container">
        """

        # Add each snapshot for this test
        for snapshot_name in test.snapshot_names:
            # Find the snapshot file
            snapshot_file = None
            for path, info in snapshots.items():
                if info.test_info and info.test_info.file_path == test.file_path and info.test_info.function_name == test.function_name:
                    if os.path.splitext(os.path.basename(path))[0] == snapshot_name:
                        snapshot_file = path
                        break

            if snapshot_file:
                snapshot_info = snapshots[snapshot_file]
                is_modified = snapshot_info.is_modified
                has_original = is_modified and snapshot_info.original_content is not None

                snapshot_id = f"snapshot_{test.function_name}_{snapshot_name}".replace('.', '_')

                buttons = f"""<div class="button-pair"><button id="toggle-original-{snapshot_id}" class="toggle-version-btn" onclick="toggleVersion('{snapshot_id}', 'original')">Original</button><button id="toggle-current-{snapshot_id}" class="toggle-version-btn active" onclick="toggleVersion('{snapshot_id}', 'current')">Current</button></div>"""

                html += f"""
                <div class="snapshot">
                    <div class="snapshot-header{' modified' if is_modified else ''}">
                        {snapshot_name}{' (Modified)' if is_modified else ''}
                        {buttons if has_original else ''}
                    </div>
                """

                # If the snapshot is modified and we have the original content, add toggle button
                if has_original:
                    html += f"""

                    <div id="{snapshot_id}_original" class="snapshot-content original-content">
                        {html_escape(snapshot_info.original_content)}
                    </div>
                    <div id="{snapshot_id}_current" class="snapshot-content">
                        {html_escape(snapshot_info.content)}
                    </div>
                    """
                else:
                    # Just show the current content
                    html += f"""
                    <div class="snapshot-content">{html_escape(snapshot_info.content)}</div>
                    """

                html += """
                </div>
                """

        html += """
                </div>
            </div>
        </div>
        """

    html += """
            </div>
        </div>
        <div class="fixed-nav">
            <button id="prevButton" class="nav-button">Previous</button>
            <button id="nextButton" class="nav-button">Next</button>
        </div>
        <script>
            document.getElementById('search').addEventListener('input', function(e) {
                const searchTerm = e.target.value.toLowerCase();
                const tests = document.querySelectorAll('.test-container');

                tests.forEach(test => {
                    const testName = test.getAttribute('data-test').toLowerCase();
                    const fileName = test.getAttribute('data-file').toLowerCase();
                    const content = test.textContent.toLowerCase();

                    if (testName.includes(searchTerm) || fileName.includes(searchTerm) || content.includes(searchTerm)) {
                        test.style.display = '';
                    } else {
                        test.style.display = 'none';
                    }
                });
            });

            // Navigation functionality
            const snapshots = document.querySelectorAll('.snapshot');
            let currentIndex = -1;

            function navigateToSnapshot(index) {
                if (index >= 0 && index < snapshots.length) {
                    if (currentIndex >= 0 && currentIndex < snapshots.length) {
                        snapshots[currentIndex].style.border = '1px solid #eee';
                    }

                    currentIndex = index;
                    const snapshot = snapshots[currentIndex];
                    snapshot.style.border = '2px solid #4CAF50';
                    snapshot.scrollIntoView({ behavior: 'smooth', block: 'center' });
                }

                // Update button states
                const prevButton = document.getElementById('prevButton');
                if (prevButton) {
                    prevButton.disabled = currentIndex <= 0;
                }
                const nextButton = document.getElementById('nextButton');
                if (nextButton) {
                    nextButton.disabled = currentIndex >= snapshots.length - 1;
                }
            }

            function navigateNext() {
                navigateToSnapshot(currentIndex + 1);
            }

            function navigatePrevious() {
                navigateToSnapshot(currentIndex - 1);
            }

            // Initialize navigation
            document.getElementById('prevButton')?.addEventListener('click', navigatePrevious);
            document.getElementById('nextButton')?.addEventListener('click', navigateNext);

            // Start at the first snapshot
            if (snapshots.length > 0) {
                navigateToSnapshot(0);
            } else {
                const prevButton = document.getElementById('prevButton');
                if (prevButton) {
                    prevButton.disabled = true;
                }
                const nextButton = document.getElementById('nextButton');
                if (nextButton) {
                    nextButton.disabled = true;
                }
            }

            // Toggle between original and current version
            function toggleVersion(snapshotId, version) {
                const originalElement = document.getElementById(snapshotId + '_original');
                const currentElement = document.getElementById(snapshotId + '_current');

                if (version === 'original') {
                    originalElement.style.display = 'block';
                    currentElement.style.display = 'none';
                } else {
                    originalElement.style.display = 'none';
                    currentElement.style.display = 'block';
                }
                const button = document.getElementById("toggle-" + version + "-" + snapshotId);
                const otherButton = document.getElementById("toggle-" + (version === 'original' ? 'current' : 'original') + "-" + snapshotId);
                button.classList.add('active');
                otherButton.classList.remove('active');
            }
        </script>
    </body>
    </html>
    """

    # Write the HTML to the output file
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(html)

def mark_modified_snapshots(snapshots):
    """Mark snapshots that are modified according to git."""
    try:
        # Get both staged and unstaged changes
        staged = subprocess.check_output(['git', 'diff', '--name-only', '--cached']).decode('utf-8').splitlines()
        unstaged = subprocess.check_output(['git', 'diff', '--name-only']).decode('utf-8').splitlines()
        untracked = subprocess.check_output(['git', 'ls-files', '--others', '--exclude-standard']).decode('utf-8').splitlines()

        # Combine all changes
        modified_files = set(staged + unstaged + untracked)

        # Mark modified snapshots
        for name, info in snapshots.items():
            info.is_modified = info.file_path in modified_files

        return True
    except subprocess.CalledProcessError:
        return False

def load_snapshot_registry():
    """Load the snapshot registry from the .pytest_snapshot_data directory."""
    registry_path = os.path.join(os.getcwd(), ".pytest_snapshot_data", "registry.json")
    print(f"Looking for registry at: {registry_path}")

    if not os.path.exists(registry_path):
        print(f"Registry file not found at {registry_path}")
        return None

    try:
        with open(registry_path, 'r') as f:
            registry = json.load(f)

        print(f"Loaded registry with {len(registry)} test entries")

        return registry
    except Exception as e:
        print(f"Error loading registry: {e}")
        return None

def extract_test_info_from_registry(registry):
    """Extract test information from the registry."""
    tests = []
    for test_id, test_data in registry.items():
        # Skip tests with no snapshots
        if not test_data["snapshots"]:
            continue

        # Read the test code from the file
        try:
            with open(test_data["file_path"], 'r') as f:
                file_content = f.read()

            # Extract the function code - get the whole function
            line_number = test_data["line_number"]
            lines = file_content.splitlines()

            # Find the end of the function
            function_lines = []
            indent_level = None
            for i in range(line_number - 1, len(lines)):
                line = lines[i]

                # Skip empty lines at the beginning
                if not line.strip() and not function_lines:
                    continue

                # Determine the indentation level of the function
                if indent_level is None and line.strip():
                    indent_level = len(line) - len(line.lstrip())

                # Add the line to the function
                function_lines.append(line)

                # Check if we've reached the end of the function
                if i + 1 < len(lines):
                    next_line = lines[i + 1]
                    if next_line.strip() and len(next_line) - len(next_line.lstrip()) <= cast(int, indent_level):
                        break

            function_code = '\n'.join(function_lines)

            # Create a TestInfo object with normalized snapshot paths
            snapshot_names = []
            for snapshot_path in test_data["snapshots"]:
                # Extract just the filename without extension
                snapshot_file = os.path.basename(snapshot_path)
                snapshot_name = os.path.splitext(snapshot_file)[0]
                snapshot_names.append(snapshot_name)

            test = TestInfo(
                file_path=test_data["file_path"],
                function_name=test_data["function_name"],
                class_name=test_data["class_name"],
                code=function_code,
                line_number=test_data["line_number"],
                snapshot_names=snapshot_names
            )
            tests.append(test)
        except Exception as e:
            print(f"Error extracting test info for {test_id}: {e}")

    return tests

def match_tests_with_snapshots_from_registry(tests, snapshots, registry):
    """Match tests with snapshots using the registry."""
    matched_tests = []
    matched_snapshots = {}

    # Create a mapping of snapshot filenames to their full paths
    snapshot_filename_map = {}
    for snapshot_file, info in snapshots.items():
        filename = os.path.basename(snapshot_file)
        snapshot_filename_map[filename] = snapshot_file

    # For each test, find its snapshots
    for test in tests:
        has_snapshot = False

        # Get the test's registry entry
        test_id = None
        for tid, data in registry.items():
            if (data["function_name"] == test.function_name and
                data["file_path"] == test.file_path and
                data["class_name"] == test.class_name):
                test_id = tid
                break

        if test_id:
            # Get the snapshots for this test
            for snapshot_path in registry[test_id]["snapshots"]:
                snapshot_file = os.path.basename(snapshot_path)

                # Find the snapshot in our snapshots dict
                if snapshot_file in snapshot_filename_map:
                    full_path = snapshot_filename_map[snapshot_file]
                    has_snapshot = True
                    snapshots[full_path].test_info = test
                    matched_snapshots[full_path] = snapshots[full_path]

        if has_snapshot:
            matched_tests.append(test)

    return matched_tests, matched_snapshots

def process_snapshots_from_registry(registry, test_dir):
    """Process snapshots using the registry information directly."""
    all_tests = []
    matched_snapshots = {}

    # First, find all snapshot files
    all_snapshots = {}
    for root, _, files in os.walk(test_dir):
        for file in files:
            if file.endswith('.snapshot'):
                file_path = os.path.join(root, file)
                try:
                    with open(file_path, 'r', encoding='utf-8') as f:
                        content = f.read()
                    all_snapshots[file_path] = SnapshotInfo(file_path=file_path, content=content)
                except Exception as e:
                    print(f"Error reading snapshot file {file_path}: {e}")

    # Process each test in the registry
    for test_id, test_data in registry.items():
        # Skip tests with no snapshots
        if not test_data["snapshots"]:
            continue

        # Extract test information
        try:
            with open(test_data["file_path"], 'r') as f:
                file_content = f.read()

            # Extract the function code
            line_number = test_data["line_number"]
            lines = file_content.splitlines()

            # Find the end of the function
            function_lines = []
            indent_level = None
            for i in range(line_number - 1, len(lines)):
                line = lines[i]

                # Skip empty lines at the beginning
                if not line.strip() and not function_lines:
                    continue

                # Determine the indentation level of the function
                if indent_level is None and line.strip():
                    indent_level = len(line) - len(line.lstrip())

                # Add the line to the function
                function_lines.append(line)

                # Check if we've reached the end of the function
                if i + 1 < len(lines):
                    next_line = lines[i + 1]
                    if next_line.strip() and len(next_line) - len(next_line.lstrip()) <= cast(int, indent_level):
                        break

            function_code = '\n'.join(function_lines)

            # Create a TestInfo object
            test = TestInfo(
                file_path=test_data["file_path"],
                function_name=test_data["function_name"],
                class_name=test_data["class_name"],
                code=function_code,
                line_number=test_data["line_number"],
                snapshot_names=[os.path.splitext(os.path.basename(s))[0] for s in test_data["snapshots"]]
            )

            # Add the test to our list
            all_tests.append(test)

            # Match the test with its snapshots
            for snapshot_path in test_data["snapshots"]:
                # Normalize the path to handle different path formats
                normalized_path = os.path.normpath(snapshot_path)

                # Try to find the snapshot in all_snapshots
                if normalized_path in all_snapshots:
                    all_snapshots[normalized_path].test_info = test
                    matched_snapshots[normalized_path] = all_snapshots[normalized_path]
                else:
                    # Try to match by basename
                    basename = os.path.basename(normalized_path)
                    for path, info in all_snapshots.items():
                        if os.path.basename(path) == basename:
                            info.test_info = test
                            matched_snapshots[path] = info
                            break

        except Exception as e:
            print(f"Error processing test {test_id}: {e}")

    return all_tests, matched_snapshots

def generate_report_from_registry(registry, snapshot_files, output_path):
    """Generate an HTML report using the registry information."""
    print("Generating report from registry...")

    def from_repo(file_path):
        return "".join(file_path.split("/relationalai-python/")[1])

    # Load all snapshot files
    all_snapshots = {}
    for file_path in snapshot_files:
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            all_snapshots[file_path] = SnapshotInfo(file_path=from_repo(file_path), content=content)
            # Also add with normalized path
            norm_path = os.path.normpath(file_path)
            all_snapshots[norm_path] = SnapshotInfo(file_path=from_repo(file_path), content=content)
        except Exception as e:
            print(f"Error reading snapshot file {file_path}: {e}")

    print(f"Loaded {len(snapshot_files)} snapshot files")
    # Mark modified snapshots and get original content
    mark_modified_snapshots(all_snapshots)
    for path, info in all_snapshots.items():
        if info.is_modified:
            info.original_content = get_original_snapshot_content(info.file_path)

    # Process each test in the registry
    all_tests = []
    matched_snapshots = {}

    for test_id, test_data in registry.items():
        # Skip tests with no snapshots
        if not test_data.get("snapshots") or len(test_data["snapshots"]) == 0:
            continue

        # Extract test information
        try:
            with open(test_data["file_path"], 'r') as f:
                file_content = f.read()

            # Extract the function code
            line_number = test_data["line_number"]
            lines = file_content.splitlines()

            # Find the end of the function
            function_lines = []
            indent_level = None
            for i in range(line_number - 1, len(lines)):
                line = lines[i]

                # Skip empty lines at the beginning
                if not line.strip() and not function_lines:
                    continue

                # Determine the indentation level of the function
                if indent_level is None and line.strip():
                    indent_level = len(line) - len(line.lstrip())

                # Add the line to the function
                function_lines.append(line)

                # Check if we've reached the end of the function
                if i + 1 < len(lines):
                    next_line = lines[i + 1]
                    if next_line.strip() and len(next_line) - len(next_line.lstrip()) <= cast(int, indent_level):
                        break

            function_code = '\n'.join(function_lines)

            # Create a TestInfo object
            test = TestInfo(
                file_path=test_data["file_path"],
                function_name=test_data["function_name"],
                class_name=test_data.get("class_name"),
                code=function_code,
                line_number=line_number,
                snapshot_names=[]
            )

            # Process each snapshot for this test
            found_snapshots = False
            for snapshot_path in test_data["snapshots"]:
                # Check if this snapshot is in our filtered list
                norm_path = os.path.normpath(snapshot_path)

                if norm_path in all_snapshots:
                    # Add the snapshot to the test
                    name_without_ext = os.path.splitext(os.path.basename(norm_path))[0]
                    test.snapshot_names.append(name_without_ext)
                    all_snapshots[norm_path].test_info = test
                    matched_snapshots[norm_path] = all_snapshots[norm_path]
                    found_snapshots = True

            # If we found snapshots, add the test
            if found_snapshots:
                all_tests.append(test)

        except Exception as e:
            print(f"Error processing test {test_id}: {e}")

    print(f"Found {len(all_tests)} tests with {len(matched_snapshots)} snapshots")

    # Generate HTML report
    generate_html_report(all_tests, matched_snapshots, output_path)

def html_escape(text):
    """Escape HTML special characters."""
    if text is None:
        return ""
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

def get_original_snapshot_content(snapshot_path):
    """Get the original content of a snapshot from Git."""
    try:
        # Try to get the content from the last commit
        original_content = subprocess.check_output(
            ['git', 'show', f'HEAD:{snapshot_path}'],
            stderr=subprocess.DEVNULL
        ).decode('utf-8')
        return original_content
    except subprocess.CalledProcessError:
        # File might be new and not in Git yet
        return None

def main():
    parser = argparse.ArgumentParser(description='Generate an HTML report of snapshot tests')
    parser.add_argument('--output', default=None, help='Output HTML file (defaults to a temp file)')
    parser.add_argument('--test-dir', default='tests/unit/', help='Test directory to scan')

    args = parser.parse_args()

    # Create a temporary file for the HTML report if not specified
    if not args.output:
        temp_dir = tempfile.mkdtemp()
        args.output = os.path.join(temp_dir, "snapshot_report.html")

    # Use the registry to get test information
    registry = load_snapshot_registry()
    if not registry:
        print("No snapshot registry found. Run tests first to create it.")
        return

    # Generate the report using the registry
    generate_report_from_registry(registry, args.test_dir, args.output)

    print(f"Report generated: {args.output}")

    # Open the report in the default browser
    webbrowser.open(f"file://{os.path.abspath(args.output)}")

if __name__ == '__main__':
    main()


from __future__ import annotations
import atexit
from collections import defaultdict
import json
import re
import textwrap
import uuid

from pandas import DataFrame
from typing import Any, Optional, Literal, TYPE_CHECKING
from snowflake.snowpark import Session

from v0.relationalai import debugging
from v0.relationalai.clients import result_helpers
from v0.relationalai.clients.util import IdentityParser, escape_for_f_string
from v0.relationalai.clients.resources.snowflake import APP_NAME, create_resources_instance
from v0.relationalai.semantics.metamodel import ir, executor as e, factory as f
from v0.relationalai.semantics.rel import Compiler
from v0.relationalai.clients.config import Config
from v0.relationalai.tools.constants import Generation, QUERY_ATTRIBUTES_HEADER
from v0.relationalai.tools.query_utils import prepare_metadata_for_headers

if TYPE_CHECKING:
    from v0.relationalai.semantics.snowflake import Table

class RelExecutor(e.Executor):
    """Executes Rel code using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        wide_outputs: bool = False,
        connection: Session | None = None,
        config: Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.wide_outputs = wide_outputs
        self.compiler = Compiler()
        self.connection = connection
        self.config = config or Config()
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                # NOTE: language="rel" is required for Rel execution. It is the default, but
                # we set it explicitly here to be sure.
                self._resources = create_resources_instance(
                    config=self.config,
                    dry_run=self.dry_run,
                    connection=self.connection,
                    language="rel",
                )
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True, "rel")
        return self._resources

    def check_graph_index(self, headers: dict[str, Any] | None = None):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from v0.relationalai.semantics.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine
        engine_size = self.resources.config.get_default_engine_size()

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.maybe_poll_use_index(
                app_name=app_name,
                sources=sources,
                model=model,
                engine_name=self.engine,
                engine_size=engine_size,
                program_span_id=program_span_id,
                headers=headers,
            )

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from v0.relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _export(self, raw_code: str, dest: Table, actual_cols: list[str], declared_cols: list[str], update:bool, headers: dict[str, Any] | None = None):
        # _export is Snowflake-specific and requires Snowflake Resources
        # It calls Snowflake stored procedures (APP_NAME.api.exec_into_table, etc.)
        # LocalResources doesn't support this functionality
        from v0.relationalai.clients.local import LocalResources
        if isinstance(self.resources, LocalResources):
            raise NotImplementedError("Export functionality is not supported in local mode. Use Snowflake Resources instead.")
        
        _exec = self.resources._exec
        output_table = "out" + str(uuid.uuid4()).replace("-", "_")
        txn_id = None
        artifacts = None
        dest_database, dest_schema, dest_table, _ = IdentityParser(dest._fqn, require_all_parts=True).to_list()
        dest_fqn = dest._fqn
        assert self.resources._session  # All Snowflake Resources have _session
        with debugging.span("transaction"):
            try:
                with debugging.span("exec_format") as span:
                    res = _exec(f"call {APP_NAME}.api.exec_into_table(?, ?, ?, ?, ?, NULL, ?, {headers}, ?, ?);", [self.database, self.engine, raw_code, output_table, False, True, False, None])
                    txn_id = json.loads(res[0]["EXEC_INTO_TABLE"])["rai_transaction_id"]
                    span["txn_id"] = txn_id

                with debugging.span("write_table", txn_id=txn_id):
                    out_sample = _exec(f"select * from {APP_NAME}.results.{output_table} limit 1;")
                    if out_sample:
                        keys = set([k.lower() for k in out_sample[0].as_dict().keys()])
                        fields = []
                        ix = 0
                        for name in declared_cols:
                            if name in actual_cols:
                                field = f"col{ix:03} as \"{name}\"" if f"col{ix:03}" in keys else f"NULL as {name}"
                                ix += 1
                            else:
                                field = f"NULL as \"{name}\""
                            fields.append(field)
                        names = ", ".join(fields)
                        if not update:

                            createTableLogic = f"""
                                        CREATE TABLE {dest_fqn} AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{output_table};
                            """
                            if dest._is_iceberg:
                                assert dest._iceberg_config is not None
                                external_volume_clause = ""
                                if dest._iceberg_config.external_volume:
                                    external_volume_clause = f"EXTERNAL_VOLUME = '{dest._iceberg_config.external_volume}'"
                                createTableLogic = f"""
                                            CREATE ICEBERG TABLE {dest_fqn}
                                            CATALOG = "SNOWFLAKE"
                                            {external_volume_clause}
                                            AS
                                            SELECT {names}
                                            FROM {APP_NAME}.results.{output_table};
                                """

                            _exec(f"""
                                BEGIN
                                    -- Check if table exists
                                    IF (EXISTS (
                                        SELECT 1
                                        FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                        WHERE table_schema = '{dest_schema}'
                                        AND table_name = '{dest_table}'
                                    )) THEN
                                        -- Insert into existing table
                                        EXECUTE IMMEDIATE '
                                            BEGIN
                                                TRUNCATE TABLE {dest_fqn};
                                                INSERT INTO {dest_fqn}
                                                SELECT {names}
                                                FROM {APP_NAME}.results.{output_table};
                                            END;
                                        ';
                                    ELSE
                                        -- Create table based on the SELECT
                                        EXECUTE IMMEDIATE '
                                            {escape_for_f_string(createTableLogic)}
                                        ';
                                    END IF;
                                    CALL {APP_NAME}.api.drop_result_table('{output_table}');
                                END;
                            """)
                        else:
                            _exec(f"""
                                begin
                                    INSERT INTO {dest_fqn} SELECT {names} FROM {APP_NAME}.results.{output_table};
                                    call {APP_NAME}.api.drop_result_table('{output_table}');
                                end;
                            """)
            except Exception as e:
                msg = str(e).lower()
                if "no columns returned" in msg or "columns of results could not be determined" in msg:
                    if not update:
                        # TODO: this doesn't handle a case when we have empty result, table doesn't exists and we need to create it.
                        #   To handle it we need to get mapping from the export columns like `col000` with the SQL type to a real column name.
                        _exec(f"""
                            BEGIN
                                IF (EXISTS (
                                    SELECT 1
                                    FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                    WHERE table_schema = '{dest_schema}'
                                    AND table_name = '{dest_table}'
                                )) THEN
                                    EXECUTE IMMEDIATE 'TRUNCATE TABLE {dest_fqn}';
                                END IF;
                            END;
                        """)
                else:
                    raise e
            if txn_id:
                # These methods are available on all Snowflake Resources
                artifact_info = self.resources._list_exec_async_artifacts(txn_id, headers=headers)
                with debugging.span("fetch"):
                    artifacts = self.resources._download_results(artifact_info, txn_id, "ABORTED")
            return artifacts

    def execute(self, model: ir.Model, task: ir.Task, format: Literal["pandas", "snowpark"] = "pandas",
                export_to: Optional[Table] = None, update: bool = False, meta: dict[str, Any] | None = None) -> Any:
        # Format meta as headers
        json_meta = prepare_metadata_for_headers(meta)
        headers = {QUERY_ATTRIBUTES_HEADER: json_meta} if json_meta else {}

        self.check_graph_index(headers)
        resources= self.resources

        rules_code = ""
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                install_span["compile_type"] = "model"
                base = textwrap.dedent("""
                    declare pyrel_error_attrs(err in ::std::common::UInt128, attr in ::std::common::String, v) requires true

                """)
                rules_code = base + self.compiler.compile(model, {"wide_outputs": self.wide_outputs})
                install_span["rel"] = rules_code
                rules_code = resources.create_models_code([("pyrel_qb_0", rules_code)])
                self._last_model = model


        with debugging.span("compile", metamodel=task) as compile_span:
            compile_span["compile_type"] = "query"
            base = textwrap.dedent("""
                def output(:pyrel_error, err, attr, val):
                    pyrel_error_attrs(err, attr, val)

            """)
            task_model = f.compute_model(f.logical([task]))
            task_code, task_model = self.compiler.compile_inner(task_model, {"no_declares": True, "wide_outputs": self.wide_outputs})
            task_code = base + task_code
            compile_span["rel"] = task_code

        full_code = textwrap.dedent(f"""
            {rules_code}
            {task_code}
        """)

        if self.dry_run:
            return DataFrame()

        cols, extra_cols = self._compute_cols(task, task_model)

        if not export_to:
            if format == "pandas":
                raw_results = resources.exec_raw(self.database, self.engine, full_code, False, nowait_durable=True, headers=headers)
                df, errs = result_helpers.format_results(raw_results, None, cols, generation=Generation.QB)  # Pass None for task parameter
                self.report_errors(errs)
                # Rename columns if wide outputs is enabled
                if self.wide_outputs and len(cols) - len(extra_cols) == len(df.columns):
                    df.columns = cols[: len(df.columns)]

                return self._postprocess_df(self.config, df, extra_cols)
            elif format == "snowpark":
                results, raw = resources.exec_format(self.database, self.engine, full_code, cols, format=format, readonly=False, nowait_durable=True, headers=headers)
                if raw:
                    df, errs = result_helpers.format_results(raw, None, cols, generation=Generation.QB)  # Pass None for task parameter
                    self.report_errors(errs)

                return results
        else:
            assert cols
            # The result cols should be a superset of the actual cols.
            result_cols = export_to._col_names
            if result_cols is not None:
                assert all(col in result_cols or col in extra_cols for col in cols)
            else:
                result_cols = [col for col in cols if col not in extra_cols]
            assert result_cols
            raw = self._export(full_code, export_to, cols, result_cols, update, headers)
            errors = []
            if raw:
                dataframe, errors = result_helpers.format_results(raw, None, result_cols, generation=Generation.QB)
                self.report_errors(errors)
            return DataFrame()

    # NOTE(coey): this is added temporarily to support executing Rel for the solvers library in EA.
    # It can be removed once this is no longer needed by the solvers library.
    def execute_raw(self, raw_rel:str, readonly:bool=True, query_timeout_mins:int|None=None) -> DataFrame:
        raw_results = self.resources.exec_raw(self.database, self.engine, raw_rel, readonly, nowait_durable=True, query_timeout_mins=query_timeout_mins)
        df, errs = result_helpers.format_results(raw_results, None, generation=Generation.QB)  # Pass None for task parameter
        self.report_errors(errs)
        return df



def humanized_duration(ms: int) -> str:
    """Format duration in milliseconds to human-readable format with spaces.

    Converts milliseconds to a human-readable string with up to 3 time components
    (days, hours, minutes, seconds) separated by spaces.

    Args:
        ms: Duration in milliseconds

    Returns:
        Formatted string like '1h 2m 3.5s' (max 3 components).
        Returns empty string for zero duration.

    Examples:
        >>> humanized_duration(1000)
        '1.0s'
        >>> humanized_duration(90000)
        '1m 30.0s'
        >>> humanized_duration(3665000)
        '1h 1m 5.0s'
        >>> humanized_duration(95400000)
        '1d 2h 30m'
    """
    # Convert milliseconds to other time units
    seconds = ms / 1000
    minutes = seconds / 60
    hours = minutes / 60
    days = hours / 24

    # Extract display components using modulo to get remainders
    display_days = int(days)
    display_hours = int(hours % 24)
    display_minutes = int(minutes % 60)
    display_seconds = seconds % 60

    # Build list of non-zero time components
    parts = []
    if display_days > 0:
        parts.append(f"{display_days}d")
    if display_hours > 0:
        parts.append(f"{display_hours}h")
    if display_minutes > 0:
        parts.append(f"{display_minutes}m")
    if display_seconds > 0:
        # Show seconds with one decimal place (no thousand separators for time)
        parts.append(f"{display_seconds:.1f}s")

    # Return up to 3 components, space-separated
    return " ".join(parts[:3])


def format_duration(duration_seconds: float) -> str:
    """Format duration in seconds to human-readable format.

    Args:
        duration_seconds: Duration in seconds

    Returns:
        Formatted string like '500ms', '2.5s', '1m30s', '1h5m'

    Examples:
        >>> format_duration(0.5)
        '500ms'
        >>> format_duration(2.5)
        '2.5s'
        >>> format_duration(90)
        '1m30s'
        >>> format_duration(3665)
        '1h1m'
    """
    # For durations less than 1 second, show milliseconds
    if duration_seconds < 1:
        milliseconds = round(duration_seconds * 1000)
        # Handle edge case where rounding gives us 1000ms
        if milliseconds >= 1000:
            return "1.0s"
        return f"{milliseconds}ms"

    # For durations less than 1 minute, show seconds with one decimal
    elif duration_seconds < 60:
        return f"{round(duration_seconds, 1)}s"

    # For durations less than 1 hour, show minutes and seconds
    elif duration_seconds < 3600:
        minutes = int(duration_seconds // 60)
        seconds = int(duration_seconds % 60)
        if seconds == 0:
            return f"{minutes}m"
        else:
            return f"{minutes}m{seconds}s"
    else:
        hours = int(duration_seconds // 3600)
        minutes = int((duration_seconds % 3600) // 60)
        if minutes == 0:
            return f"{hours}h"
        else:
            return f"{hours}h{minutes}m"

def humanized_bytes(bytes: int) -> str:
    """Format bytes to human-readable format."""
    if bytes < 1024:
        return f"{bytes} B"
    elif bytes < 1024 ** 2:
        return f"{bytes / 1024:.1f} KB"
    elif bytes < 1024 ** 3:
        return f"{bytes / 1024 ** 2:.1f} MB"
    elif bytes < 1024 ** 4:
        return f"{bytes / 1024 ** 3:.1f} GB"
    else:
        return f"{bytes / 1024 ** 4:.1f} TB"

def str_to_bool(s):
    # Adapted from snowflake's convert_str_to_bool
    if s is None:
        return None
    if s.lower() in ("true", "t", "yes", "y", "on", "1"):
        return True
    if s.lower() in ("false", "f", "no", "n", "off", "0"):
        return False
    raise ValueError(f"Invalid boolean value: {s}")

def default_serialize(obj):
    return '<skipped>'
